(function () {
  'use strict';

  function _classCallCheck(instance, Constructor) {
    if (!(instance instanceof Constructor)) {
      throw new TypeError("Cannot call a class as a function");
    }
  }

  function _defineProperties(target, props) {
    for (var i = 0; i < props.length; i++) {
      var descriptor = props[i];
      descriptor.enumerable = descriptor.enumerable || false;
      descriptor.configurable = true;
      if ("value" in descriptor) descriptor.writable = true;
      Object.defineProperty(target, descriptor.key, descriptor);
    }
  }

  function _createClass(Constructor, protoProps, staticProps) {
    if (protoProps) _defineProperties(Constructor.prototype, protoProps);
    if (staticProps) _defineProperties(Constructor, staticProps);
    return Constructor;
  }

  /**
   * Checks if the device is online by performing a network health check to the learning analytics platform
   * @return {Promise} A promise that resolves to true if the device is online, or false if it's offline
   */
  function isOnline() {
    // if the browser knows it's not connected, don't try the health check at all
    if (window.navigator.onLine === false) {
      return Promise.resolve(false);
    }

    var request = new window.XMLHttpRequest();
    var HEALTH_CHECK_URL = 'https://s3.amazonaws.com/tortuga-static/health-check.json';
    return new Promise(function (resolve) {
      // when the request loads, attempt to parse the response text as JSON. It could be that the user is on a LAN which
      // causes network requests to succeed, but respond with a "login to our wifi" page (like at PBS or at most coffee
      // shops).
      request.addEventListener('load', function (event) {
        var responseText = event.target.responseText;

        try {
          JSON.parse(responseText);
          resolve(true);
        } catch (e) {
          resolve(false);
        }
      });
      request.addEventListener('error', function () {
        resolve(false);
      });
      request.open('GET', HEALTH_CHECK_URL);
      request.send();
    });
  }

  function t(t, e) {
    if (!(t instanceof e)) throw new TypeError("Cannot call a class as a function");
  }

  function e(t, e) {
    for (var i = 0; i < e.length; i++) {
      var r = e[i];
      r.enumerable = r.enumerable || !1, r.configurable = !0, "value" in r && (r.writable = !0), Object.defineProperty(t, r.key, r);
    }
  }

  function i(t, i, r) {
    return i && e(t.prototype, i), r && e(t, r), t;
  }

  var r = "undefined" != typeof window ? window : "undefined" != typeof global ? global : "undefined" != typeof self ? self : {};

  function n(t, e) {
    return t(e = {
      exports: {}
    }, e.exports), e.exports;
  }

  var o = n(function (t, e) {
    t.exports = function () {
      function t(t) {
        if (this.app = "loki", void 0 !== t && (this.app = t), this.catalog = null, !this.checkAvailability()) throw new Error("indexedDB does not seem to be supported for your environment");
      }

      function e(t) {
        this.db = null, this.initializeLokiCatalog(t);
      }

      return t.prototype.checkAvailability = function () {
        return !("undefined" == typeof indexedDB || !indexedDB);
      }, t.prototype.loadDatabase = function (t, i) {
        var r = this.app,
            n = this;
        null !== this.catalog && null !== this.catalog.db ? this.catalog.getAppKey(r, t, function (t) {
          if ("function" == typeof i) {
            if (0 === t.id) return void i(null);
            i(t.val);
          } else console.log(t.val);
        }) : this.catalog = new e(function (e) {
          n.catalog = e, n.loadDatabase(t, i);
        });
      }, t.prototype.loadKey = t.prototype.loadDatabase, t.prototype.saveDatabase = function (t, i, r) {
        var n = this.app,
            o = this;

        function s(t) {
          t && !0 === t.success ? r(null) : r(new Error("Error saving database"));
        }

        null !== this.catalog && null !== this.catalog.db ? this.catalog.setAppKey(n, t, i, s) : this.catalog = new e(function (e) {
          o.catalog = e, e.setAppKey(n, t, i, s);
        });
      }, t.prototype.saveKey = t.prototype.saveDatabase, t.prototype.deleteDatabase = function (t, i) {
        var r = this.app,
            n = this;
        null !== this.catalog && null !== this.catalog.db ? this.catalog.getAppKey(r, t, function (t) {
          var e = t.id;
          0 !== e ? n.catalog.deleteAppKey(e, i) : "function" == typeof i && i({
            success: !0
          });
        }) : this.catalog = new e(function (e) {
          n.catalog = e, n.deleteDatabase(t, i);
        });
      }, t.prototype.deleteKey = t.prototype.deleteDatabase, t.prototype.deleteDatabasePartitions = function (t) {
        var e = this;
        this.getDatabaseList(function (i) {
          i.forEach(function (i) {
            i.startsWith(t) && e.deleteDatabase(i);
          });
        });
      }, t.prototype.getDatabaseList = function (t) {
        var i = this.app,
            r = this;
        null !== this.catalog && null !== this.catalog.db ? this.catalog.getAppKeys(i, function (e) {
          for (var i = [], r = 0; r < e.length; r++) i.push(e[r].key);

          "function" == typeof t ? t(i) : i.forEach(function (t) {
            console.log(t);
          });
        }) : this.catalog = new e(function (e) {
          r.catalog = e, r.getDatabaseList(t);
        });
      }, t.prototype.getKeyList = t.prototype.getDatabaseList, t.prototype.getCatalogSummary = function (t) {
        this.app;
        var i = this;
        null !== this.catalog && null !== this.catalog.db ? this.catalog.getAllKeys(function (e) {
          for (var i, r, n, o, s, a = [], l = 0; l < e.length; l++) i = e[l], n = i.app || "", o = i.key || "", s = i.val || "", r = 2 * n.length + 2 * o.length + s.length + 1, a.push({
            app: i.app,
            key: i.key,
            size: r
          });

          "function" == typeof t ? t(a) : a.forEach(function (t) {
            console.log(t);
          });
        }) : this.catalog = new e(function (e) {
          i.catalog = e, i.getCatalogSummary(t);
        });
      }, e.prototype.initializeLokiCatalog = function (t) {
        var e = indexedDB.open("LokiCatalog", 1),
            i = this;
        e.onupgradeneeded = function (t) {
          var e = t.target.result;

          if (e.objectStoreNames.contains("LokiAKV") && e.deleteObjectStore("LokiAKV"), !e.objectStoreNames.contains("LokiAKV")) {
            var i = e.createObjectStore("LokiAKV", {
              keyPath: "id",
              autoIncrement: !0
            });
            i.createIndex("app", "app", {
              unique: !1
            }), i.createIndex("key", "key", {
              unique: !1
            }), i.createIndex("appkey", "appkey", {
              unique: !0
            });
          }
        }, e.onsuccess = function (e) {
          i.db = e.target.result, "function" == typeof t && t(i);
        }, e.onerror = function (t) {
          throw t;
        };
      }, e.prototype.getAppKey = function (t, e, i) {
        var r,
            n = this.db.transaction(["LokiAKV"], "readonly"),
            o = n.objectStore("LokiAKV"),
            s = o.index("appkey"),
            a = t + "," + e,
            l = s.get(a);
        l.onsuccess = (r = i, function (t) {
          var e = t.target.result;
          null == e && (e = {
            id: 0,
            success: !1
          }), "function" == typeof r ? r(e) : console.log(e);
        }), l.onerror = function (t) {
          return function (e) {
            if ("function" != typeof t) throw e;
            t({
              id: 0,
              success: !1
            });
          };
        }(i);
      }, e.prototype.getAppKeyById = function (t, e, i) {
        var r = this.db.transaction(["LokiAKV"], "readonly"),
            n = r.objectStore("LokiAKV"),
            o = n.get(t);

        o.onsuccess = function (t, e) {
          return function (i) {
            "function" == typeof e ? e(i.target.result, t) : console.log(i.target.result);
          };
        }(i, e);
      }, e.prototype.setAppKey = function (t, e, i, r) {
        var n,
            o = this.db.transaction(["LokiAKV"], "readwrite"),
            s = o.objectStore("LokiAKV"),
            a = s.index("appkey"),
            l = t + "," + e,
            h = a.get(l);
        h.onsuccess = function (n) {
          var o = n.target.result;
          null == o ? o = {
            app: t,
            key: e,
            appkey: t + "," + e,
            val: i
          } : o.val = i;
          var a,
              l = s.put(o);
          l.onerror = (a = r, function (t) {
            "function" == typeof a ? a({
              success: !1
            }) : (console.error("LokiCatalog.setAppKey (set) onerror"), console.error(h.error));
          }), l.onsuccess = function (t) {
            return function (e) {
              "function" == typeof t && t({
                success: !0
              });
            };
          }(r);
        }, h.onerror = (n = r, function (t) {
          "function" == typeof n ? n({
            success: !1
          }) : (console.error("LokiCatalog.setAppKey (get) onerror"), console.error(h.error));
        });
      }, e.prototype.deleteAppKey = function (t, e) {
        var i,
            r = this.db.transaction(["LokiAKV"], "readwrite"),
            n = r.objectStore("LokiAKV"),
            o = n.delete(t);
        o.onsuccess = (i = e, function (t) {
          "function" == typeof i && i({
            success: !0
          });
        }), o.onerror = function (t) {
          return function (e) {
            "function" == typeof t ? t({
              success: !1
            }) : (console.error("LokiCatalog.deleteAppKey raised onerror"), console.error(o.error));
          };
        }(e);
      }, e.prototype.getAppKeys = function (t, e) {
        var i,
            r = this.db.transaction(["LokiAKV"], "readonly"),
            n = r.objectStore("LokiAKV"),
            o = n.index("app"),
            s = IDBKeyRange.only(t),
            a = o.openCursor(s);
        a.onsuccess = function (t, e) {
          return function (i) {
            var r = i.target.result;

            if (r) {
              var n = r.value;
              t.push(n), r.continue();
            } else "function" == typeof e ? e(t) : console.log(t);
          };
        }([], e), a.onerror = (i = e, function (t) {
          "function" == typeof i ? i(null) : (console.error("LokiCatalog.getAppKeys raised onerror"), console.error(t));
        });
      }, e.prototype.getAllKeys = function (t) {
        var e,
            i = this.db.transaction(["LokiAKV"], "readonly"),
            r = i.objectStore("LokiAKV"),
            n = r.openCursor();
        n.onsuccess = function (t, e) {
          return function (i) {
            var r = i.target.result;

            if (r) {
              var n = r.value;
              t.push(n), r.continue();
            } else "function" == typeof e ? e(t) : console.log(t);
          };
        }([], t), n.onerror = (e = t, function (t) {
          "function" == typeof e && e(null);
        });
      }, t;
    }();
  }),
      s = n(function (t, e) {
    t.exports = function () {
      var t = Object.prototype.hasOwnProperty,
          e = {
        copyProperties: function (t, e) {
          var i;

          for (i in t) e[i] = t[i];
        },
        resolveTransformObject: function (t, i, r) {
          var n, o;
          if ("number" != typeof r && (r = 0), ++r >= 10) return t;

          for (n in t) "string" == typeof t[n] && 0 === t[n].indexOf("[%lktxp]") ? (o = t[n].substring(8), i.hasOwnProperty(o) && (t[n] = i[o])) : "object" == typeof t[n] && (t[n] = e.resolveTransformObject(t[n], i, r));

          return t;
        },
        resolveTransformParams: function (t, i) {
          var r,
              n,
              o = [];
          if (void 0 === i) return t;

          for (r = 0; r < t.length; r++) n = f(t[r], "shallow-recurse-objects"), o.push(e.resolveTransformObject(n, i));

          return o;
        }
      },
          i = {
        aeq: n,
        lt: s,
        gt: a
      };

      function n(t, e) {
        var i, r, n, o;
        if (t === e) return !0;

        if (!t || !e || !0 === t || !0 === e || t != t || e != e) {
          switch (t) {
            case void 0:
            case null:
              n = 1;
              break;

            case !1:
              n = 3;
              break;

            case !0:
              n = 4;
              break;

            case "":
              n = 5;
              break;

            default:
              n = t == t ? 9 : 0;
          }

          switch (e) {
            case void 0:
            case null:
              o = 1;
              break;

            case !1:
              o = 3;
              break;

            case !0:
              o = 4;
              break;

            case "":
              o = 5;
              break;

            default:
              o = e == e ? 9 : 0;
          }

          if (9 !== n || 9 !== o) return n === o;
        }

        return i = Number(t), r = Number(e), i == i || r == r ? i === r : (i = t.toString(), r = e.toString(), i == r);
      }

      function s(t, e, i) {
        var r, n, o, s;

        if (!t || !e || !0 === t || !0 === e || t != t || e != e) {
          switch (t) {
            case void 0:
            case null:
              o = 1;
              break;

            case !1:
              o = 3;
              break;

            case !0:
              o = 4;
              break;

            case "":
              o = 5;
              break;

            default:
              o = t == t ? 9 : 0;
          }

          switch (e) {
            case void 0:
            case null:
              s = 1;
              break;

            case !1:
              s = 3;
              break;

            case !0:
              s = 4;
              break;

            case "":
              s = 5;
              break;

            default:
              s = e == e ? 9 : 0;
          }

          if (9 !== o || 9 !== s) return o === s ? i : o < s;
        }

        return r = Number(t), n = Number(e), r == r && n == n ? r < n || !(r > n) && i : r == r && n != n || (n != n || r == r) && (t < e || !(t > e) && (t == e ? i : (r = t.toString(), n = e.toString(), r < n || r == n && i)));
      }

      function a(t, e, i) {
        var r, n, o, s;

        if (!t || !e || !0 === t || !0 === e || t != t || e != e) {
          switch (t) {
            case void 0:
            case null:
              o = 1;
              break;

            case !1:
              o = 3;
              break;

            case !0:
              o = 4;
              break;

            case "":
              o = 5;
              break;

            default:
              o = t == t ? 9 : 0;
          }

          switch (e) {
            case void 0:
            case null:
              s = 1;
              break;

            case !1:
              s = 3;
              break;

            case !0:
              s = 4;
              break;

            case "":
              s = 5;
              break;

            default:
              s = e == e ? 9 : 0;
          }

          if (9 !== o || 9 !== s) return o === s ? i : o > s;
        }

        return r = Number(t), n = Number(e), r == r && n == n ? r > n || !(r < n) && i : (r != r || n == n) && (n == n && r != r || t > e || !(t < e) && (t == e ? i : (r = t.toString(), n = e.toString(), r > n || r == n && i)));
      }

      function l(t, e, r) {
        return i.aeq(t, e) ? 0 : i.lt(t, e, !1) ? r ? 1 : -1 : i.gt(t, e, !1) ? r ? -1 : 1 : 0;
      }

      function h(e, i, r, n, o) {
        var s = o || 0,
            a = i[s];
        if (null == e || !t.call(e, a)) return !1;
        var l = !1,
            c = e[a];
        if (s + 1 >= i.length) l = r(c, n);else if (Array.isArray(c)) for (var u = 0, d = c.length; u < d && !0 !== (l = h(c[u], i, r, n, s + 1)); u += 1);else l = h(c, i, r, n, s + 1);
        return l;
      }

      function c(e) {
        return "string" == typeof e || Array.isArray(e) ? function (t) {
          return -1 !== e.indexOf(t);
        } : "object" == typeof e && null !== e ? function (i) {
          return t.call(e, i);
        } : null;
      }

      function u(e, i) {
        for (var r in i) if (t.call(i, r)) return d[r](e, i[r]);

        return !1;
      }

      var d = {
        $eq: function (t, e) {
          return t === e;
        },
        $aeq: function (t, e) {
          return t == e;
        },
        $ne: function (t, e) {
          return e != e ? t == t : t !== e;
        },
        $dteq: function (t, e) {
          return i.aeq(t, e);
        },
        $gt: function (t, e) {
          return i.gt(t, e, !1);
        },
        $gte: function (t, e) {
          return i.gt(t, e, !0);
        },
        $lt: function (t, e) {
          return i.lt(t, e, !1);
        },
        $lte: function (t, e) {
          return i.lt(t, e, !0);
        },
        $jgt: function (t, e) {
          return t > e;
        },
        $jgte: function (t, e) {
          return t >= e;
        },
        $jlt: function (t, e) {
          return t < e;
        },
        $jlte: function (t, e) {
          return t <= e;
        },
        $between: function (t, e) {
          return null != t && i.gt(t, e[0], !0) && i.lt(t, e[1], !0);
        },
        $jbetween: function (t, e) {
          return null != t && t >= e[0] && t <= e[1];
        },
        $in: function (t, e) {
          return -1 !== e.indexOf(t);
        },
        $nin: function (t, e) {
          return -1 === e.indexOf(t);
        },
        $keyin: function (t, e) {
          return t in e;
        },
        $nkeyin: function (t, e) {
          return !(t in e);
        },
        $definedin: function (t, e) {
          return void 0 !== e[t];
        },
        $undefinedin: function (t, e) {
          return void 0 === e[t];
        },
        $regex: function (t, e) {
          return e.test(t);
        },
        $containsString: function (t, e) {
          return "string" == typeof t && -1 !== t.indexOf(e);
        },
        $containsNone: function (t, e) {
          return !d.$containsAny(t, e);
        },
        $containsAny: function (t, e) {
          var i = c(t);
          return null !== i && (Array.isArray(e) ? e.some(i) : i(e));
        },
        $contains: function (t, e) {
          var i = c(t);
          return null !== i && (Array.isArray(e) ? e.every(i) : i(e));
        },
        $type: function (t, e) {
          var i = typeof t;
          return "object" === i && (Array.isArray(t) ? i = "array" : t instanceof Date && (i = "date")), "object" != typeof e ? i === e : u(i, e);
        },
        $finite: function (t, e) {
          return e === isFinite(t);
        },
        $size: function (t, e) {
          return !!Array.isArray(t) && ("object" != typeof e ? t.length === e : u(t.length, e));
        },
        $len: function (t, e) {
          return "string" == typeof t && ("object" != typeof e ? t.length === e : u(t.length, e));
        },
        $where: function (t, e) {
          return !0 === e(t);
        },
        $not: function (t, e) {
          return !u(t, e);
        },
        $and: function (t, e) {
          for (var i = 0, r = e.length; i < r; i += 1) if (!u(t, e[i])) return !1;

          return !0;
        },
        $or: function (t, e) {
          for (var i = 0, r = e.length; i < r; i += 1) if (u(t, e[i])) return !0;

          return !1;
        }
      },
          p = {
        $eq: d.$eq,
        $aeq: !0,
        $dteq: !0,
        $gt: !0,
        $gte: !0,
        $lt: !0,
        $lte: !0,
        $in: !0,
        $between: !0
      };

      function f(t, e) {
        if (null == t) return null;
        var i,
            r = e || "parse-stringify";

        switch (r) {
          case "parse-stringify":
            i = JSON.parse(JSON.stringify(t));
            break;

          case "jquery-extend-deep":
            i = jQuery.extend(!0, {}, t);
            break;

          case "shallow":
            i = Object.create(t.constructor.prototype), Object.keys(t).map(function (e) {
              i[e] = t[e];
            });
            break;

          case "shallow-assign":
            i = Object.create(t.constructor.prototype), Object.assign(i, t);
            break;

          case "shallow-recurse-objects":
            i = f(t, "shallow");
            var n = Object.keys(t);
            n.forEach(function (e) {
              "object" == typeof t[e] && "Object" === t[e].constructor.name && (i[e] = f(t[e], "shallow-recurse-objects"));
            });
        }

        return i;
      }

      function y() {
        try {
          return window && void 0 !== window.localStorage && null !== window.localStorage;
        } catch (t) {
          return !1;
        }
      }

      function v() {}

      function g(t, e) {
        this.filename = t || "loki.db", this.collections = [], this.databaseVersion = 1.5, this.engineVersion = 1.5, this.autosave = !1, this.autosaveInterval = 5e3, this.autosaveHandle = null, this.throttledSaves = !0, this.options = {}, this.persistenceMethod = null, this.persistenceAdapter = null, this.throttledSavePending = !1, this.throttledCallbacks = [], this.verbose = !(!e || !e.hasOwnProperty("verbose")) && e.verbose, this.events = {
          init: [],
          loaded: [],
          flushChanges: [],
          close: [],
          changes: [],
          warning: []
        }, e && e.hasOwnProperty("env") ? this.ENV = e.env : this.ENV = void 0 !== r && (r.android || r.NSObject) ? "NATIVESCRIPT" : "undefined" == typeof window ? "NODEJS" : void 0 !== r && r.window && process ? "NODEJS" : "undefined" != typeof document ? -1 === document.URL.indexOf("http://") && -1 === document.URL.indexOf("https://") ? "CORDOVA" : "BROWSER" : "CORDOVA", "undefined" === this.ENV && (this.ENV = "NODEJS"), this.configureOptions(e, !0), this.on("init", this.clearChanges);
      }

      function m(t) {
        this.hashStore = {}, this.options = t || {}, this.options.hasOwnProperty("asyncResponses") || (this.options.asyncResponses = !1), this.options.hasOwnProperty("asyncTimeout") || (this.options.asyncTimeout = 50);
      }

      function b(t, e) {
        if (this.mode = "reference", this.adapter = null, this.options = e || {}, this.dbref = null, this.dbname = "", this.pageIterator = {}, !t) throw new Error("LokiPartitioningAdapter requires a (non-reference mode) adapter on construction");
        if ("reference" === t.mode) throw new Error("LokiPartitioningAdapter cannot be instantiated with a reference mode adapter");
        this.adapter = t, this.options.hasOwnProperty("paging") || (this.options.paging = !1), this.options.hasOwnProperty("pageSize") || (this.options.pageSize = 26214400), this.options.hasOwnProperty("delimiter") || (this.options.delimiter = "$<\n");
      }

      function w() {
        try {
          this.fs = require("fs");
        } catch (t) {
          this.fs = null;
        }
      }

      function I() {}

      function O(t, e) {
        return e = e || {}, this.collection = t, this.filteredrows = [], this.filterInitialized = !1, this;
      }

      function x(t, e, i) {
        this.collection = t, this.name = e, this.rebuildPending = !1, this.options = i || {}, this.options.hasOwnProperty("persistent") || (this.options.persistent = !1), this.options.hasOwnProperty("sortPriority") || (this.options.sortPriority = "passive"), this.options.hasOwnProperty("minRebuildInterval") || (this.options.minRebuildInterval = 1), this.resultset = new O(t), this.resultdata = [], this.resultsdirty = !1, this.cachedresultset = null, this.filterPipeline = [], this.sortFunction = null, this.sortCriteria = null, this.sortCriteriaSimple = null, this.sortDirty = !1, this.events = {
          rebuild: []
        };
      }

      function k(e, i) {
        this.name = e, this.data = [], this.idIndex = [], this.binaryIndices = {}, this.constraints = {
          unique: {},
          exact: {}
        }, this.uniqueNames = [], this.transforms = {}, this.objType = e, this.dirty = !0, this.cachedIndex = null, this.cachedBinaryIndex = null, this.cachedData = null;
        var r = this;
        (i = i || {}).hasOwnProperty("unique") && (Array.isArray(i.unique) || (i.unique = [i.unique]), i.unique.forEach(function (t) {
          r.uniqueNames.push(t), r.constraints.unique[t] = new N(t);
        })), i.hasOwnProperty("exact") && i.exact.forEach(function (t) {
          r.constraints.exact[t] = new M(t);
        }), this.adaptiveBinaryIndices = !i.hasOwnProperty("adaptiveBinaryIndices") || i.adaptiveBinaryIndices, this.transactional = !!i.hasOwnProperty("transactional") && i.transactional, this.cloneObjects = !!i.hasOwnProperty("clone") && i.clone, this.cloneMethod = i.hasOwnProperty("cloneMethod") ? i.cloneMethod : "parse-stringify", this.asyncListeners = !!i.hasOwnProperty("asyncListeners") && i.asyncListeners, this.disableMeta = !!i.hasOwnProperty("disableMeta") && i.disableMeta, this.disableChangesApi = !i.hasOwnProperty("disableChangesApi") || i.disableChangesApi, this.disableDeltaChangesApi = !i.hasOwnProperty("disableDeltaChangesApi") || i.disableDeltaChangesApi, this.disableChangesApi && (this.disableDeltaChangesApi = !0), this.autoupdate = !!i.hasOwnProperty("autoupdate") && i.autoupdate, this.serializableIndices = !i.hasOwnProperty("serializableIndices") || i.serializableIndices, this.ttl = {
          age: null,
          ttlInterval: null,
          daemon: null
        }, this.setTTL(i.ttl || -1, i.ttlInterval), this.maxId = 0, this.DynamicViews = [], this.events = {
          insert: [],
          update: [],
          "pre-insert": [],
          "pre-update": [],
          close: [],
          flushbuffer: [],
          error: [],
          delete: [],
          warning: []
        }, this.changes = [], this.ensureId();
        var n = [];
        if (i && i.indices) if ("[object Array]" === Object.prototype.toString.call(i.indices)) n = i.indices;else {
          if ("string" != typeof i.indices) throw new TypeError("Indices needs to be a string or an array of strings");
          n = [i.indices];
        }

        for (var o = 0; o < n.length; o++) this.ensureIndex(n[o]);

        function s(t, e) {
          var i = null !== e && "object" == typeof e ? Object.keys(e) : null;

          if (i && i.length && ["string", "boolean", "number"].indexOf(typeof e) < 0) {
            for (var n = {}, o = 0; o < i.length; o++) {
              var a = i[o];
              if (e.hasOwnProperty(a)) if (!t.hasOwnProperty(a) || r.uniqueNames.indexOf(a) >= 0 || "$loki" == a || "meta" == a) n[a] = e[a];else {
                var l = s(t[a], e[a]);
                void 0 !== l && l != {} && (n[a] = l);
              }
            }

            return 0 === Object.keys(n).length ? void 0 : n;
          }

          return t === e ? void 0 : e;
        }

        function a() {
          r.changes = [];
        }

        this.observerCallback = function (e) {
          var i = "function" == typeof Set ? new Set() : [];
          i.add || (i.add = function (t) {
            return -1 === this.indexOf(t) && this.push(t), this;
          }), e.forEach(function (t) {
            i.add(t.object);
          }), i.forEach(function (e) {
            if (!t.call(e, "$loki")) return r.removeAutoUpdateObserver(e);

            try {
              r.update(e);
            } catch (t) {}
          });
        }, this.getChangeDelta = function (t, e) {
          return e ? s(e, t) : JSON.parse(JSON.stringify(t));
        }, this.getObjectDelta = s, this.getChanges = function () {
          return r.changes;
        }, this.flushChanges = a, this.setChangesApi = function (t) {
          r.disableChangesApi = !t, t || (r.disableDeltaChangesApi = !1);
        }, this.on("delete", function (t) {
          r.disableChangesApi || r.createChange(r.name, "R", t);
        }), this.on("warning", function (t) {
          r.console.warn(t);
        }), a();
      }

      function D(t) {
        return -1 !== t.indexOf(".");
      }

      function S(t) {
        return parseFloat(t, 10);
      }

      function P(t, e) {
        return t + e;
      }

      function A(t, e) {
        return t - e;
      }

      function C(t) {
        return t.reduce(P, 0) / t.length;
      }

      function q(t, e, i) {
        if (!1 === i) return t[e];

        for (var r = e.split("."), n = t; r.length > 0;) n = n[r.shift()];

        return n;
      }

      function $(t, e, i) {
        for (var r, n, o = 0, s = t.length; o < s;) {
          if (n = o + s >> 1, 0 === (r = i.apply(null, [e, t[n]]))) return {
            found: !0,
            index: n
          };
          r < 0 ? s = n : o = n + 1;
        }

        return {
          found: !1,
          index: s
        };
      }

      function E(t) {
        return function (e, i) {
          return $(e, i, t);
        };
      }

      function j() {}

      function N(t) {
        this.field = t, this.keyMap = {}, this.lokiMap = {};
      }

      function M(t) {
        this.index = {}, this.field = t;
      }

      return v.prototype.events = {}, v.prototype.asyncListeners = !1, v.prototype.on = function (t, e) {
        var i,
            r = this;
        return Array.isArray(t) ? (t.forEach(function (t) {
          r.on(t, e);
        }), e) : ((i = this.events[t]) || (i = this.events[t] = []), i.push(e), e);
      }, v.prototype.emit = function (t) {
        var e = this,
            i = Array.prototype.slice.call(arguments, 1);
        if (!t || !this.events[t]) throw new Error("No event " + t + " defined");
        this.events[t].forEach(function (t) {
          e.asyncListeners ? setTimeout(function () {
            t.apply(e, i);
          }, 1) : t.apply(e, i);
        });
      }, v.prototype.addListener = v.prototype.on, v.prototype.removeListener = function (t, e) {
        var i = this;
        if (Array.isArray(t)) t.forEach(function (t) {
          i.removeListener(t, e);
        });else if (this.events[t]) {
          var r = this.events[t];
          r.splice(r.indexOf(e), 1);
        }
      }, g.prototype = new v(), g.prototype.constructor = g, g.prototype.getIndexedAdapter = function () {
        var t;
        return "function" == typeof require && (t = o), t;
      }, g.prototype.configureOptions = function (t, e) {
        var i = {
          fs: w,
          localStorage: I,
          memory: m
        };

        if (this.options = {}, this.persistenceMethod = null, this.persistenceAdapter = null, void 0 !== t) {
          if (this.options = t, this.options.hasOwnProperty("persistenceMethod") && "function" == typeof i[t.persistenceMethod] && (this.persistenceMethod = t.persistenceMethod, this.persistenceAdapter = new i[t.persistenceMethod]()), this.options.hasOwnProperty("adapter") && (this.persistenceMethod = "adapter", this.persistenceAdapter = t.adapter, this.options.adapter = null), t.autoload && e) {
            var r = this;
            setTimeout(function () {
              r.loadDatabase(t, t.autoloadCallback);
            }, 1);
          }

          this.options.hasOwnProperty("autosaveInterval") && (this.autosaveDisable(), this.autosaveInterval = parseInt(this.options.autosaveInterval, 10)), this.options.hasOwnProperty("autosave") && this.options.autosave && (this.autosaveDisable(), this.autosave = !0, this.options.hasOwnProperty("autosaveCallback") ? this.autosaveEnable(t, t.autosaveCallback) : this.autosaveEnable()), this.options.hasOwnProperty("throttledSaves") && (this.throttledSaves = this.options.throttledSaves);
        }

        this.options.hasOwnProperty("serializationMethod") || (this.options.serializationMethod = "normal"), this.options.hasOwnProperty("destructureDelimiter") || (this.options.destructureDelimiter = "$<\n"), null === this.persistenceAdapter && (this.persistenceMethod = {
          NODEJS: "fs",
          BROWSER: "localStorage",
          CORDOVA: "localStorage",
          MEMORY: "memory"
        }[this.ENV], this.persistenceMethod && (this.persistenceAdapter = new i[this.persistenceMethod]()));
      }, g.prototype.copy = function (t) {
        var e,
            i,
            r = new g(this.filename, {
          env: "NA"
        });
        if (t = t || {}, r.loadJSONObject(this, {
          retainDirtyFlags: !0
        }), t.hasOwnProperty("removeNonSerializable") && !0 === t.removeNonSerializable) for (r.autosaveHandle = null, r.persistenceAdapter = null, e = r.collections.length, i = 0; i < e; i++) r.collections[i].constraints = null, r.collections[i].ttl = null;
        return r;
      }, g.prototype.addCollection = function (t, e) {
        var i,
            r = this.collections.length;

        if (e && !0 === e.disableMeta) {
          if (!1 === e.disableChangesApi) throw new Error("disableMeta option cannot be passed as true when disableChangesApi is passed as false");
          if (!1 === e.disableDeltaChangesApi) throw new Error("disableMeta option cannot be passed as true when disableDeltaChangesApi is passed as false");
          if ("number" == typeof e.ttl && e.ttl > 0) throw new Error("disableMeta option cannot be passed as true when ttl is enabled");
        }

        for (i = 0; i < r; i += 1) if (this.collections[i].name === t) return this.collections[i];

        var n = new k(t, e);
        return this.collections.push(n), this.verbose && (n.console = console), n;
      }, g.prototype.loadCollection = function (t) {
        if (!t.name) throw new Error("Collection must have a name property to be loaded");
        this.collections.push(t);
      }, g.prototype.getCollection = function (t) {
        var e,
            i = this.collections.length;

        for (e = 0; e < i; e += 1) if (this.collections[e].name === t) return this.collections[e];

        return this.emit("warning", "collection " + t + " not found"), null;
      }, g.prototype.renameCollection = function (t, e) {
        var i = this.getCollection(t);
        return i && (i.name = e), i;
      }, g.prototype.listCollections = function () {
        for (var t = this.collections.length, e = []; t--;) e.push({
          name: this.collections[t].name,
          type: this.collections[t].objType,
          count: this.collections[t].data.length
        });

        return e;
      }, g.prototype.removeCollection = function (t) {
        var e,
            i = this.collections.length;

        for (e = 0; e < i; e += 1) if (this.collections[e].name === t) {
          var r = new k(t, {}),
              n = this.collections[e];

          for (var o in n) n.hasOwnProperty(o) && r.hasOwnProperty(o) && (n[o] = r[o]);

          return void this.collections.splice(e, 1);
        }
      }, g.prototype.getName = function () {
        return this.name;
      }, g.prototype.serializeReplacer = function (t, e) {
        switch (t) {
          case "autosaveHandle":
          case "persistenceAdapter":
          case "constraints":
          case "ttl":
            return null;

          case "throttledSavePending":
          case "throttledCallbacks":
            return;

          default:
            return e;
        }
      }, g.prototype.serialize = function (t) {
        switch ((t = t || {}).hasOwnProperty("serializationMethod") || (t.serializationMethod = this.options.serializationMethod), t.serializationMethod) {
          case "normal":
            return JSON.stringify(this, this.serializeReplacer);

          case "pretty":
            return JSON.stringify(this, this.serializeReplacer, 2);

          case "destructured":
            return this.serializeDestructured();

          default:
            return JSON.stringify(this, this.serializeReplacer);
        }
      }, g.prototype.toJson = g.prototype.serialize, g.prototype.serializeDestructured = function (t) {
        var e,
            i,
            r,
            n,
            o,
            s = [];
        if ((t = t || {}).hasOwnProperty("partitioned") || (t.partitioned = !1), t.hasOwnProperty("delimited") || (t.delimited = !0), t.hasOwnProperty("delimiter") || (t.delimiter = this.options.destructureDelimiter), !0 === t.partitioned && t.hasOwnProperty("partition") && t.partition >= 0) return this.serializeCollection({
          delimited: t.delimited,
          delimiter: t.delimiter,
          collectionIndex: t.partition
        });

        for ((o = new g(this.filename)).loadJSONObject(this), e = 0; e < o.collections.length; e++) o.collections[e].data = [];

        if (!0 === t.partitioned && -1 === t.partition) return o.serialize({
          serializationMethod: "normal"
        });

        for (s.push(o.serialize({
          serializationMethod: "normal"
        })), o = null, e = 0; e < this.collections.length; e++) if (r = this.serializeCollection({
          delimited: t.delimited,
          delimiter: t.delimiter,
          collectionIndex: e
        }), !1 === t.partitioned && !1 === t.delimited) {
          if (!Array.isArray(r)) throw new Error("a nondelimited, non partitioned collection serialization did not return an expected array");

          for (n = r.length, i = 0; i < n; i++) s.push(r[i]), r[i] = null;

          s.push("");
        } else s.push(r);

        return t.partitioned ? (t.delimited, s) : t.delimited ? (s.push(""), s.join(t.delimiter)) : (s.push(""), s);
      }, g.prototype.serializeCollection = function (t) {
        var e,
            i,
            r = [];
        if ((t = t || {}).hasOwnProperty("delimited") || (t.delimited = !0), !t.hasOwnProperty("collectionIndex")) throw new Error("serializeCollection called without 'collectionIndex' option");

        for (e = this.collections[t.collectionIndex].data.length, r = [], i = 0; i < e; i++) r.push(JSON.stringify(this.collections[t.collectionIndex].data[i]));

        return t.delimited ? (r.push(""), r.join(t.delimiter)) : r;
      }, g.prototype.deserializeDestructured = function (t, e) {
        var i,
            r,
            n,
            o = [],
            s = 0,
            a = 1,
            l = !1;

        if ((e = e || {}).hasOwnProperty("partitioned") || (e.partitioned = !1), e.hasOwnProperty("delimited") || (e.delimited = !0), e.hasOwnProperty("delimiter") || (e.delimiter = this.options.destructureDelimiter), e.partitioned) {
          if (e.hasOwnProperty("partition")) return -1 === e.partition ? i = JSON.parse(t[0]) : this.deserializeCollection(t[e.partition + 1], e);

          for (i = JSON.parse(t[0]), r = i.collections.length, s = 0; s < r; s++) i.collections[s].data = this.deserializeCollection(t[s + 1], e);

          return i;
        }

        if (e.delimited) {
          if (o = t.split(e.delimiter), t = null, 0 === o.length) return null;
        } else o = t;

        for (i = JSON.parse(o[0]), r = i.collections.length, o[0] = null; !l;) o[a], "" === o[a] ? ++s > r && (l = !0) : (n = JSON.parse(o[a]), i.collections[s].data.push(n)), o[a++] = null;

        return i;
      }, g.prototype.deserializeCollection = function (t, e) {
        var i,
            r,
            n = [];

        for ((e = e || {}).hasOwnProperty("partitioned") || (e.partitioned = !1), e.hasOwnProperty("delimited") || (e.delimited = !0), e.hasOwnProperty("delimiter") || (e.delimiter = this.options.destructureDelimiter), e.delimited ? (n = t.split(e.delimiter)).pop() : n = t, r = n.length, i = 0; i < r; i++) n[i] = JSON.parse(n[i]);

        return n;
      }, g.prototype.loadJSON = function (t, e) {
        var i;
        if (0 === t.length) i = {};else switch (this.options.serializationMethod) {
          case "normal":
          case "pretty":
            i = JSON.parse(t);
            break;

          case "destructured":
            i = this.deserializeDestructured(t);
            break;

          default:
            i = JSON.parse(t);
        }
        this.loadJSONObject(i, e);
      }, g.prototype.loadJSONObject = function (t, i) {
        var r,
            n,
            o,
            s,
            a,
            l,
            h = 0,
            c = t.collections ? t.collections.length : 0;

        function u(t) {
          var r,
              n = i[t.name];
          return n.proto ? (r = n.inflate || e.copyProperties, function (t) {
            var e = new n.proto();
            return r(t, e), e;
          }) : n.inflate;
        }

        for (this.name = t.name, t.hasOwnProperty("throttledSaves") && i && !i.hasOwnProperty("throttledSaves") && (this.throttledSaves = t.throttledSaves), this.collections = []; h < c; h += 1) {
          if (r = t.collections[h], (n = this.addCollection(r.name, {
            disableChangesApi: r.disableChangesApi,
            disableDeltaChangesApi: r.disableDeltaChangesApi,
            disableMeta: r.disableMeta
          })).adaptiveBinaryIndices = !!r.hasOwnProperty("adaptiveBinaryIndices") && !0 === r.adaptiveBinaryIndices, n.transactional = r.transactional, n.asyncListeners = r.asyncListeners, n.cloneObjects = r.cloneObjects, n.cloneMethod = r.cloneMethod || "parse-stringify", n.autoupdate = r.autoupdate, n.changes = r.changes, i && !0 === i.retainDirtyFlags ? n.dirty = r.dirty : n.dirty = !1, o = r.data.length, s = 0, i && i.hasOwnProperty(r.name)) for (a = u(r); s < o; s++) l = a(r.data[s]), n.data[s] = l, n.addAutoUpdateObserver(l);else for (; s < o; s++) n.data[s] = r.data[s], n.addAutoUpdateObserver(n.data[s]);
          if (n.maxId = void 0 === r.maxId ? 0 : r.maxId, n.idIndex = r.idIndex, void 0 !== r.binaryIndices && (n.binaryIndices = r.binaryIndices), void 0 !== r.transforms && (n.transforms = r.transforms), n.ensureId(), n.uniqueNames = [], r.hasOwnProperty("uniqueNames")) for (n.uniqueNames = r.uniqueNames, s = 0; s < n.uniqueNames.length; s++) n.ensureUniqueIndex(n.uniqueNames[s]);

          if (void 0 !== r.DynamicViews) {
            for (var d = 0; d < r.DynamicViews.length; d++) {
              var p = r.DynamicViews[d],
                  f = n.addDynamicView(p.name, p.options);
              f.resultdata = p.resultdata, f.resultsdirty = p.resultsdirty, f.filterPipeline = p.filterPipeline, f.sortCriteria = p.sortCriteria, f.sortFunction = null, f.sortDirty = p.sortDirty, f.resultset.filteredrows = p.resultset.filteredrows, f.resultset.filterInitialized = p.resultset.filterInitialized, f.rematerialize({
                removeWhereFilters: !0
              });
            }

            t.databaseVersion < 1.5 && (n.ensureAllIndexes(!0), n.dirty = !0);
          }
        }
      }, g.prototype.close = function (t) {
        this.autosave && (this.autosaveDisable(), this.autosaveDirty() && (this.saveDatabase(t), t = void 0)), t && this.on("close", t), this.emit("close");
      }, g.prototype.generateChangesNotification = function (t) {
        function e(t) {
          return t.name;
        }

        var i = [],
            r = t || this.collections.map(e);
        return this.collections.forEach(function (t) {
          -1 !== r.indexOf(e(t)) && (i = i.concat(t.getChanges()));
        }), i;
      }, g.prototype.serializeChanges = function (t) {
        return JSON.stringify(this.generateChangesNotification(t));
      }, g.prototype.clearChanges = function () {
        this.collections.forEach(function (t) {
          t.flushChanges && t.flushChanges();
        });
      }, m.prototype.loadDatabase = function (t, e) {
        var i = this;
        this.options.asyncResponses ? setTimeout(function () {
          i.hashStore.hasOwnProperty(t) ? e(i.hashStore[t].value) : e(null);
        }, this.options.asyncTimeout) : this.hashStore.hasOwnProperty(t) ? e(this.hashStore[t].value) : e(null);
      }, m.prototype.saveDatabase = function (t, e, i) {
        var r,
            n = this;
        this.options.asyncResponses ? setTimeout(function () {
          r = n.hashStore.hasOwnProperty(t) ? n.hashStore[t].savecount : 0, n.hashStore[t] = {
            savecount: r + 1,
            lastsave: new Date(),
            value: e
          }, i();
        }, this.options.asyncTimeout) : (r = this.hashStore.hasOwnProperty(t) ? this.hashStore[t].savecount : 0, this.hashStore[t] = {
          savecount: r + 1,
          lastsave: new Date(),
          value: e
        }, i());
      }, m.prototype.deleteDatabase = function (t, e) {
        this.hashStore.hasOwnProperty(t) && delete this.hashStore[t], "function" == typeof e && e();
      }, b.prototype.loadDatabase = function (t, e) {
        var i = this;
        this.dbname = t, this.dbref = new g(t), this.adapter.loadDatabase(t, function (t) {
          if (t) {
            "string" != typeof t && e(new Error("LokiPartitioningAdapter received an unexpected response from inner adapter loadDatabase()"));
            var r = JSON.parse(t);
            i.dbref.loadJSONObject(r), r = null, i.dbref.collections.length, 0 !== i.dbref.collections.length ? (i.pageIterator = {
              collection: 0,
              pageIndex: 0
            }, i.loadNextPartition(0, function () {
              e(i.dbref);
            })) : e(i.dbref);
          } else e(t);
        });
      }, b.prototype.loadNextPartition = function (t, e) {
        var i = this.dbname + "." + t,
            r = this;
        if (!0 === this.options.paging) return this.pageIterator.pageIndex = 0, void this.loadNextPage(e);
        this.adapter.loadDatabase(i, function (i) {
          var n = r.dbref.deserializeCollection(i, {
            delimited: !0,
            collectionIndex: t
          });
          r.dbref.collections[t].data = n, ++t < r.dbref.collections.length ? r.loadNextPartition(t, e) : e();
        });
      }, b.prototype.loadNextPage = function (t) {
        var e = this.dbname + "." + this.pageIterator.collection + "." + this.pageIterator.pageIndex,
            i = this;
        this.adapter.loadDatabase(e, function (e) {
          var r = e.split(i.options.delimiter);
          e = "";
          var n,
              o = r.length,
              s = "" === r[o - 1];

          for (s && (r.pop(), o = r.length, "" === r[o - 1] && 1 === o && (r.pop(), o = r.length)), n = 0; n < o; n++) i.dbref.collections[i.pageIterator.collection].data.push(JSON.parse(r[n])), r[n] = null;

          r = [], s ? ++i.pageIterator.collection < i.dbref.collections.length ? i.loadNextPartition(i.pageIterator.collection, t) : t() : (i.pageIterator.pageIndex++, i.loadNextPage(t));
        });
      }, b.prototype.exportDatabase = function (t, e, i) {
        var r,
            n = e.collections.length;

        for (this.dbref = e, this.dbname = t, this.dirtyPartitions = [-1], r = 0; r < n; r++) e.collections[r].dirty && this.dirtyPartitions.push(r);

        this.saveNextPartition(function (t) {
          i(t);
        });
      }, b.prototype.saveNextPartition = function (t) {
        var e = this,
            i = this.dirtyPartitions.shift(),
            r = this.dbname + (-1 === i ? "" : "." + i);
        if (this.options.paging && -1 !== i) return this.pageIterator = {
          collection: i,
          docIndex: 0,
          pageIndex: 0
        }, void this.saveNextPage(function (i) {
          0 === e.dirtyPartitions.length ? t(i) : e.saveNextPartition(t);
        });
        var n = this.dbref.serializeDestructured({
          partitioned: !0,
          delimited: !0,
          partition: i
        });
        this.adapter.saveDatabase(r, n, function (i) {
          i ? t(i) : 0 === e.dirtyPartitions.length ? t(null) : e.saveNextPartition(t);
        });
      }, b.prototype.saveNextPage = function (t) {
        var e = this,
            i = this.dbref.collections[this.pageIterator.collection],
            r = this.dbname + "." + this.pageIterator.collection + "." + this.pageIterator.pageIndex,
            n = 0,
            o = i.data.length,
            s = this.options.delimiter.length,
            a = "",
            l = "",
            h = !1,
            c = !1,
            u = function (i) {
          l = "", i && t(i), h ? t(null) : (e.pageIterator.pageIndex++, e.saveNextPage(t));
        };

        for (0 === i.data.length && (h = !0);;) if (h || (a = JSON.stringify(i.data[this.pageIterator.docIndex]), l += a, n += a.length, ++this.pageIterator.docIndex >= o && (h = !0)), n >= this.options.pageSize && (c = !0), c && !h || (l += this.options.delimiter, n += s), h || c) return void this.adapter.saveDatabase(r, l, u);
      }, w.prototype.loadDatabase = function (t, e) {
        var i = this;
        this.fs.stat(t, function (r, n) {
          !r && n.isFile() ? i.fs.readFile(t, {
            encoding: "utf8"
          }, function (t, i) {
            e(t ? new Error(t) : i);
          }) : e(null);
        });
      }, w.prototype.saveDatabase = function (t, e, i) {
        var r = this,
            n = t + "~";
        this.fs.writeFile(n, e, function (e) {
          e ? i(new Error(e)) : r.fs.rename(n, t, i);
        });
      }, w.prototype.deleteDatabase = function (t, e) {
        this.fs.unlink(t, function (t) {
          t ? e(new Error(t)) : e();
        });
      }, I.prototype.loadDatabase = function (t, e) {
        y() ? e(localStorage.getItem(t)) : e(new Error("localStorage is not available"));
      }, I.prototype.saveDatabase = function (t, e, i) {
        y() ? (localStorage.setItem(t, e), i(null)) : i(new Error("localStorage is not available"));
      }, I.prototype.deleteDatabase = function (t, e) {
        y() ? (localStorage.removeItem(t), e(null)) : e(new Error("localStorage is not available"));
      }, g.prototype.throttledSaveDrain = function (t, e) {
        var i = this,
            r = new Date().getTime();

        if (this.throttledSaves || t(!0), (e = e || {}).hasOwnProperty("recursiveWait") || (e.recursiveWait = !0), e.hasOwnProperty("recursiveWaitLimit") || (e.recursiveWaitLimit = !1), e.hasOwnProperty("recursiveWaitLimitDuration") || (e.recursiveWaitLimitDuration = 2e3), e.hasOwnProperty("started") || (e.started = new Date().getTime()), this.throttledSaves && this.throttledSavePending) {
          if (!e.recursiveWait) return void this.throttledCallbacks.push(t);
          this.throttledCallbacks.push(function () {
            return i.throttledSavePending ? e.recursiveWaitLimit && r - e.started > e.recursiveWaitLimitDuration ? void t(!1) : void i.throttledSaveDrain(t, e) : void t(!0);
          });
        } else t(!0);
      }, g.prototype.loadDatabaseInternal = function (t, e) {
        var i = e || function (t, e) {
          if (t) throw t;
        },
            r = this;

        null !== this.persistenceAdapter ? this.persistenceAdapter.loadDatabase(this.filename, function (e) {
          if ("string" == typeof e) {
            var n = !1;

            try {
              r.loadJSON(e, t || {}), n = !0;
            } catch (t) {
              i(t);
            }

            n && (i(null), r.emit("loaded", "database " + r.filename + " loaded"));
          } else {
            if (!e) return i(null), void r.emit("loaded", "empty database " + r.filename + " loaded");
            if (e instanceof Error) return void i(e);
            if ("object" == typeof e) return r.loadJSONObject(e, t || {}), i(null), void r.emit("loaded", "database " + r.filename + " loaded");
            i("unexpected adapter response : " + e);
          }
        }) : i(new Error("persistenceAdapter not configured"));
      }, g.prototype.loadDatabase = function (t, e) {
        var i = this;
        this.throttledSaves ? this.throttledSaveDrain(function (r) {
          if (r) return i.throttledSavePending = !0, void i.loadDatabaseInternal(t, function (t) {
            0 === i.throttledCallbacks.length ? i.throttledSavePending = !1 : i.saveDatabase(), "function" == typeof e && e(t);
          });
          "function" == typeof e && e(new Error("Unable to pause save throttling long enough to read database"));
        }, t) : this.loadDatabaseInternal(t, e);
      }, g.prototype.saveDatabaseInternal = function (t) {
        var e = t || function (t) {
          if (t) throw t;
        },
            i = this;

        null !== this.persistenceAdapter ? "reference" === this.persistenceAdapter.mode && "function" == typeof this.persistenceAdapter.exportDatabase ? this.persistenceAdapter.exportDatabase(this.filename, this.copy({
          removeNonSerializable: !0
        }), function (t) {
          i.autosaveClearFlags(), e(t);
        }) : (i.autosaveClearFlags(), this.persistenceAdapter.saveDatabase(this.filename, i.serialize(), function (t) {
          e(t);
        })) : e(new Error("persistenceAdapter not configured"));
      }, g.prototype.saveDatabase = function (t) {
        if (this.throttledSaves) {
          if (this.throttledSavePending) this.throttledCallbacks.push(t);else {
            var e = this.throttledCallbacks;
            this.throttledCallbacks = [], e.unshift(t), this.throttledSavePending = !0;
            var i = this;
            this.saveDatabaseInternal(function (t) {
              i.throttledSavePending = !1, e.forEach(function (e) {
                "function" == typeof e && setTimeout(function () {
                  e(t);
                }, 1);
              }), i.throttledCallbacks.length > 0 && i.saveDatabase();
            });
          }
        } else this.saveDatabaseInternal(t);
      }, g.prototype.save = g.prototype.saveDatabase, g.prototype.deleteDatabase = function (t, e) {
        var i = e || function (t, e) {
          if (t) throw t;
        };

        "function" != typeof t || e || (i = t), null !== this.persistenceAdapter ? this.persistenceAdapter.deleteDatabase(this.filename, function (t) {
          i(t);
        }) : i(new Error("persistenceAdapter not configured"));
      }, g.prototype.autosaveDirty = function () {
        for (var t = 0; t < this.collections.length; t++) if (this.collections[t].dirty) return !0;

        return !1;
      }, g.prototype.autosaveClearFlags = function () {
        for (var t = 0; t < this.collections.length; t++) this.collections[t].dirty = !1;
      }, g.prototype.autosaveEnable = function (t, e) {
        this.autosave = !0;
        var i = 5e3,
            r = this;
        void 0 !== this.autosaveInterval && null !== this.autosaveInterval && (i = this.autosaveInterval), this.autosaveHandle = setInterval(function () {
          r.autosaveDirty() && r.saveDatabase(e);
        }, i);
      }, g.prototype.autosaveDisable = function () {
        void 0 !== this.autosaveHandle && null !== this.autosaveHandle && (clearInterval(this.autosaveHandle), this.autosaveHandle = null);
      }, O.prototype.reset = function () {
        return this.filteredrows.length > 0 && (this.filteredrows = []), this.filterInitialized = !1, this;
      }, O.prototype.toJSON = function () {
        var t = this.copy();
        return t.collection = null, t;
      }, O.prototype.limit = function (t) {
        this.filterInitialized || 0 !== this.filteredrows.length || (this.filteredrows = this.collection.prepareFullDocIndex());
        var e = new O(this.collection);
        return e.filteredrows = this.filteredrows.slice(0, t), e.filterInitialized = !0, e;
      }, O.prototype.offset = function (t) {
        this.filterInitialized || 0 !== this.filteredrows.length || (this.filteredrows = this.collection.prepareFullDocIndex());
        var e = new O(this.collection);
        return e.filteredrows = this.filteredrows.slice(t), e.filterInitialized = !0, e;
      }, O.prototype.copy = function () {
        var t = new O(this.collection);
        return this.filteredrows.length > 0 && (t.filteredrows = this.filteredrows.slice()), t.filterInitialized = this.filterInitialized, t;
      }, O.prototype.branch = O.prototype.copy, O.prototype.transform = function (t, i) {
        var r,
            n,
            o = this;
        if ("string" == typeof t && this.collection.transforms.hasOwnProperty(t) && (t = this.collection.transforms[t]), "object" != typeof t || !Array.isArray(t)) throw new Error("Invalid transform");

        for (void 0 !== i && (t = e.resolveTransformParams(t, i)), r = 0; r < t.length; r++) switch ((n = t[r]).type) {
          case "find":
            o.find(n.value);
            break;

          case "where":
            o.where(n.value);
            break;

          case "simplesort":
            o.simplesort(n.property, n.desc || n.options);
            break;

          case "compoundsort":
            o.compoundsort(n.value);
            break;

          case "sort":
            o.sort(n.value);
            break;

          case "limit":
            o = o.limit(n.value);
            break;

          case "offset":
            o = o.offset(n.value);
            break;

          case "map":
            o = o.map(n.value, n.dataOptions);
            break;

          case "eqJoin":
            o = o.eqJoin(n.joinData, n.leftJoinKey, n.rightJoinKey, n.mapFun, n.dataOptions);
            break;

          case "mapReduce":
            o = o.mapReduce(n.mapFunction, n.reduceFunction);
            break;

          case "update":
            o.update(n.value);
            break;

          case "remove":
            o.remove();
        }

        return o;
      }, O.prototype.sort = function (t) {
        this.filterInitialized || 0 !== this.filteredrows.length || (this.filteredrows = this.collection.prepareFullDocIndex());
        var e,
            i,
            r = (e = t, i = this.collection.data, function (t, r) {
          return e(i[t], i[r]);
        });
        return this.filteredrows.sort(r), this;
      }, O.prototype.simplesort = function (t, e) {
        var i,
            r = 10,
            n = this.collection.data.length,
            o = this.filteredrows.length,
            s = this.collection.binaryIndices.hasOwnProperty(t);

        if (void 0 !== e && !1 !== e || (e = {
          desc: !1
        }), !0 === e && (e = {
          desc: !0
        }), 0 === o) {
          if (this.filterInitialized) return this;
          if (this.collection.binaryIndices.hasOwnProperty(t)) return this.collection.ensureIndex(t), this.filteredrows = this.collection.binaryIndices[t].values.slice(0), e.desc && this.filteredrows.reverse(), this;
          this.filteredrows = this.collection.prepareFullDocIndex();
        } else if (!e.disableIndexIntersect && s && (i = n / o, e.useJavascriptSorting && (r = 6), i <= r || e.forceIndexIntersect)) {
          var a,
              h = this.filteredrows,
              c = {};

          for (a = 0; a < o; a++) c[h[a]] = !0;

          var u = this.collection.binaryIndices[t].values;
          return this.filteredrows = u.filter(function (t) {
            return c[t];
          }), e.desc && this.filteredrows.reverse(), this;
        }

        if (e.useJavascriptSorting) return this.sort(function (e, i) {
          return e[t] === i[t] ? 0 : e[t] > i[t] ? 1 : e[t] < i[t] ? -1 : void 0;
        });
        var d,
            p,
            f,
            y,
            v,
            g,
            m = (d = t, p = e.desc, f = this.collection.data, function (t, e) {
          return ~d.indexOf(".") ? (g = d.split("."), y = g.reduce(function (t, e) {
            return t && t[e] || void 0;
          }, f[t]), v = g.reduce(function (t, e) {
            return t && t[e] || void 0;
          }, f[e])) : (y = f[t][d], v = f[e][d]), l(y, v, p);
        });
        return this.filteredrows.sort(m), this;
      }, O.prototype.compoundsort = function (t) {
        if (0 === t.length) throw new Error("Invalid call to compoundsort, need at least one property");
        var e;
        if (1 === t.length) return e = t[0], Array.isArray(e) ? this.simplesort(e[0], e[1]) : this.simplesort(e, !1);

        for (var i = 0, r = t.length; i < r; i += 1) e = t[i], Array.isArray(e) || (t[i] = [e, !1]);

        this.filterInitialized || 0 !== this.filteredrows.length || (this.filteredrows = this.collection.prepareFullDocIndex());
        var n,
            o,
            s = (n = t, o = this.collection.data, function (t, e) {
          return function (t, e, i) {
            for (var r, n, o, s, a, h = 0, c = 0, u = t.length; c < u; c++) if (r = t[c], ~(n = r[0]).indexOf(".") ? (a = n.split("."), o = a.reduce(function (t, e) {
              return t && t[e] || void 0;
            }, e), s = a.reduce(function (t, e) {
              return t && t[e] || void 0;
            }, i)) : (o = e[n], s = i[n]), 0 !== (h = l(o, s, r[1]))) return h;

            return 0;
          }(n, o[t], o[e]);
        });
        return this.filteredrows.sort(s), this;
      }, O.prototype.findOr = function (t) {
        for (var e = null, i = 0, r = 0, n = [], o = [], s = 0, a = this.count(), l = 0, h = t.length; l < h; l++) {
          if (e = this.branch().find(t[l]).filteredrows, (r = e.length) === a) return this;

          for (i = 0; i < r; i++) s = e[i], void 0 === o[s] && (o[s] = !0, n.push(s));
        }

        return this.filteredrows = n, this.filterInitialized = !0, this;
      }, O.prototype.$or = O.prototype.findOr, O.prototype.findAnd = function (t) {
        for (var e = 0, i = t.length; e < i; e++) {
          if (0 === this.count()) return this;
          this.find(t[e]);
        }

        return this;
      }, O.prototype.$and = O.prototype.findAnd, O.prototype.find = function (e, i) {
        if (0 === this.collection.data.length) return this.filteredrows = [], this.filterInitialized = !0, this;
        var r,
            n,
            o,
            s,
            a,
            l,
            c,
            u = e || "getAll",
            f = !1,
            y = [],
            v = [],
            g = null;

        if (i = i || !1, "object" == typeof u) {
          for (r in u) (s = {})[r] = u[r], v.push(s), t.call(u, r) && (n = r, o = u[r]);

          if (v.length > 1) return this.find({
            $and: v
          }, i);
        }

        if (!n || "getAll" === u) return i && (this.filteredrows = this.collection.data.length > 0 ? [0] : [], this.filterInitialized = !0), this;
        if ("$and" === n || "$or" === n) return this[n](o), i && this.filteredrows.length > 1 && (this.filteredrows = this.filteredrows.slice(0, 1)), this;
        if (null === o || "object" != typeof o || o instanceof Date) a = "$eq", l = o;else {
          if ("object" != typeof o) throw new Error("Do not know what you want to do.");

          for (c in o) if (t.call(o, c)) {
            a = c, l = o[c];
            break;
          }
        }
        "$regex" === a && (Array.isArray(l) ? l = new RegExp(l[0], l[1]) : l instanceof RegExp || (l = new RegExp(l)));
        var m = -1 !== n.indexOf("."),
            b = !m && !this.filterInitialized;
        b && this.collection.binaryIndices[n] && p[a] && (!0 !== this.collection.adaptiveBinaryIndices && this.collection.ensureIndex(n), f = !0, g = this.collection.binaryIndices[n]);
        var w,
            I = d[a],
            O = this.collection.data,
            x = 0,
            k = 0,
            D = 0;
        if (this.filterInitialized) {
          if (w = this.filteredrows, k = w.length, m) for (n = n.split("."), x = 0; x < k; x++) D = w[x], h(O[D], n, I, l) && y.push(D);else for (x = 0; x < k; x++) D = w[x], I(O[D][n], l) && y.push(D);
        } else if (f) {
          var S = this.collection.calculateRange(a, n, l);

          if ("$in" !== a) {
            for (x = S[0]; x <= S[1]; x++) if (!0 !== p[a]) {
              if (p[a](O[g.values[x]][n], l) && (y.push(g.values[x]), i)) return this.filteredrows = y, this.filterInitialized = !0, this;
            } else if (y.push(g.values[x]), i) return this.filteredrows = y, this.filterInitialized = !0, this;
          } else for (x = 0, k = S.length; x < k; x++) if (y.push(g.values[S[x]]), i) return this.filteredrows = y, this.filterInitialized = !0, this;
        } else if (k = O.length, m) {
          for (n = n.split("."), x = 0; x < k; x++) if (h(O[x], n, I, l) && (y.push(x), i)) return this.filteredrows = y, this.filterInitialized = !0, this;
        } else for (x = 0; x < k; x++) if (I(O[x][n], l) && (y.push(x), i)) return this.filteredrows = y, this.filterInitialized = !0, this;
        return this.filteredrows = y, this.filterInitialized = !0, this;
      }, O.prototype.where = function (t) {
        var e,
            i = [];
        if ("function" != typeof t) throw new TypeError("Argument is not a stored view or a function");
        e = t;

        try {
          if (this.filterInitialized) {
            for (var r = this.filteredrows.length; r--;) !0 === e(this.collection.data[this.filteredrows[r]]) && i.push(this.filteredrows[r]);

            return this.filteredrows = i, this;
          }

          for (var n = this.collection.data.length; n--;) !0 === e(this.collection.data[n]) && i.push(n);

          return this.filteredrows = i, this.filterInitialized = !0, this;
        } catch (t) {
          throw t;
        }
      }, O.prototype.count = function () {
        return this.filterInitialized ? this.filteredrows.length : this.collection.count();
      }, O.prototype.data = function (t) {
        var e,
            i,
            r,
            n,
            o = [],
            s = this.collection.data;

        if ((t = t || {}).removeMeta && !t.forceClones && (t.forceClones = !0, t.forceCloneMethod = t.forceCloneMethod || "shallow"), this.collection.disableDeltaChangesApi || (t.forceClones = !0, t.forceCloneMethod = "parse-stringify"), !this.filterInitialized) {
          if (0 === this.filteredrows.length) {
            if (this.collection.cloneObjects || t.forceClones) {
              for (i = s.length, n = t.forceCloneMethod || this.collection.cloneMethod, r = 0; r < i; r++) e = f(s[r], n), t.removeMeta && (delete e.$loki, delete e.meta), o.push(e);

              return o;
            }

            return s.slice();
          }

          this.filterInitialized = !0;
        }

        var a = this.filteredrows;
        if (i = a.length, this.collection.cloneObjects || t.forceClones) for (n = t.forceCloneMethod || this.collection.cloneMethod, r = 0; r < i; r++) e = f(s[a[r]], n), t.removeMeta && (delete e.$loki, delete e.meta), o.push(e);else for (r = 0; r < i; r++) o.push(s[a[r]]);
        return o;
      }, O.prototype.update = function (t) {
        if ("function" != typeof t) throw new TypeError("Argument is not a function");
        this.filterInitialized || 0 !== this.filteredrows.length || (this.filteredrows = this.collection.prepareFullDocIndex());

        for (var e, i = this.filteredrows.length, r = this.collection.data, n = 0; n < i; n++) this.collection.cloneObjects || !this.collection.disableDeltaChangesApi ? (e = f(r[this.filteredrows[n]], this.collection.cloneMethod), t(e), this.collection.update(e)) : (t(r[this.filteredrows[n]]), this.collection.update(r[this.filteredrows[n]]));

        return this;
      }, O.prototype.remove = function () {
        return this.filterInitialized || 0 !== this.filteredrows.length || (this.filteredrows = this.collection.prepareFullDocIndex()), this.collection.removeBatchByPositions(this.filteredrows), this.filteredrows = [], this;
      }, O.prototype.mapReduce = function (t, e) {
        try {
          return e(this.data().map(t));
        } catch (t) {
          throw t;
        }
      }, O.prototype.eqJoin = function (t, e, i, r, n) {
        var o,
            s,
            a,
            l = [],
            h = [],
            c = [],
            u = "function" == typeof e,
            d = "function" == typeof i,
            p = {};
        if (l = this.data(n), o = l.length, t instanceof k) h = t.chain().data(n);else if (t instanceof O) h = t.data(n);else {
          if (!Array.isArray(t)) throw new TypeError("joinData needs to be an array or result set");
          h = t;
        }
        s = h.length;

        for (var f = 0; f < s; f++) a = d ? i(h[f]) : h[f][i], p[a] = h[f];

        r || (r = function (t, e) {
          return {
            left: t,
            right: e
          };
        });

        for (var y = 0; y < o; y++) a = u ? e(l[y]) : l[y][e], c.push(r(l[y], p[a] || {}));

        return this.collection = new k("joinData"), this.collection.insert(c), this.filteredrows = [], this.filterInitialized = !1, this;
      }, O.prototype.map = function (t, e) {
        var i = this.data(e).map(t);
        return this.collection = new k("mappedData"), this.collection.insert(i), this.filteredrows = [], this.filterInitialized = !1, this;
      }, x.prototype = new v(), x.prototype.rematerialize = function (t) {
        var e, i, r;
        if (t = t || {}, this.resultdata = [], this.resultsdirty = !0, this.resultset = new O(this.collection), (this.sortFunction || this.sortCriteria || this.sortCriteriaSimple) && (this.sortDirty = !0), t.hasOwnProperty("removeWhereFilters")) for (e = this.filterPipeline.length, i = e; i--;) "where" === this.filterPipeline[i].type && (i !== this.filterPipeline.length - 1 && (this.filterPipeline[i] = this.filterPipeline[this.filterPipeline.length - 1]), this.filterPipeline.length--);
        var n = this.filterPipeline;

        for (this.filterPipeline = [], e = n.length, r = 0; r < e; r++) this.applyFind(n[r].val);

        return this.data(), this.emit("rebuild", this), this;
      }, x.prototype.branchResultset = function (t, e) {
        var i = this.resultset.branch();
        return void 0 === t ? i : i.transform(t, e);
      }, x.prototype.toJSON = function () {
        var t = new x(this.collection, this.name, this.options);
        return t.resultset = this.resultset, t.resultdata = [], t.resultsdirty = !0, t.filterPipeline = this.filterPipeline, t.sortFunction = this.sortFunction, t.sortCriteria = this.sortCriteria, t.sortCriteriaSimple = this.sortCriteriaSimple || null, t.sortDirty = this.sortDirty, t.collection = null, t;
      }, x.prototype.removeFilters = function (t) {
        t = t || {}, this.rebuildPending = !1, this.resultset.reset(), this.resultdata = [], this.resultsdirty = !0, this.cachedresultset = null, this.filterPipeline = [], this.sortFunction = null, this.sortCriteria = null, this.sortCriteriaSimple = null, this.sortDirty = !1, !0 === t.queueSortPhase && this.queueSortPhase();
      }, x.prototype.applySort = function (t) {
        return this.sortFunction = t, this.sortCriteria = null, this.sortCriteriaSimple = null, this.queueSortPhase(), this;
      }, x.prototype.applySimpleSort = function (t, e) {
        return this.sortCriteriaSimple = {
          propname: t,
          options: e || !1
        }, this.sortCriteria = null, this.sortFunction = null, this.queueSortPhase(), this;
      }, x.prototype.applySortCriteria = function (t) {
        return this.sortCriteria = t, this.sortCriteriaSimple = null, this.sortFunction = null, this.queueSortPhase(), this;
      }, x.prototype.startTransaction = function () {
        return this.cachedresultset = this.resultset.copy(), this;
      }, x.prototype.commit = function () {
        return this.cachedresultset = null, this;
      }, x.prototype.rollback = function () {
        return this.resultset = this.cachedresultset, this.options.persistent && (this.resultdata = this.resultset.data(), this.emit("rebuild", this)), this;
      }, x.prototype._indexOfFilterWithId = function (t) {
        if ("string" == typeof t || "number" == typeof t) for (var e = 0, i = this.filterPipeline.length; e < i; e += 1) if (t === this.filterPipeline[e].uid) return e;
        return -1;
      }, x.prototype._addFilter = function (t) {
        this.filterPipeline.push(t), this.resultset[t.type](t.val);
      }, x.prototype.reapplyFilters = function () {
        this.resultset.reset(), this.cachedresultset = null, this.options.persistent && (this.resultdata = [], this.resultsdirty = !0);
        var t = this.filterPipeline;
        this.filterPipeline = [];

        for (var e = 0, i = t.length; e < i; e += 1) this._addFilter(t[e]);

        return this.sortFunction || this.sortCriteria || this.sortCriteriaSimple ? this.queueSortPhase() : this.queueRebuildEvent(), this;
      }, x.prototype.applyFilter = function (t) {
        var e = this._indexOfFilterWithId(t.uid);

        return e >= 0 ? (this.filterPipeline[e] = t, this.reapplyFilters()) : (this.cachedresultset = null, this.options.persistent && (this.resultdata = [], this.resultsdirty = !0), this._addFilter(t), this.sortFunction || this.sortCriteria || this.sortCriteriaSimple ? this.queueSortPhase() : this.queueRebuildEvent(), this);
      }, x.prototype.applyFind = function (t, e) {
        return this.applyFilter({
          type: "find",
          val: t,
          uid: e
        }), this;
      }, x.prototype.applyWhere = function (t, e) {
        return this.applyFilter({
          type: "where",
          val: t,
          uid: e
        }), this;
      }, x.prototype.removeFilter = function (t) {
        var e = this._indexOfFilterWithId(t);

        if (e < 0) throw new Error("Dynamic view does not contain a filter with ID: " + t);
        return this.filterPipeline.splice(e, 1), this.reapplyFilters(), this;
      }, x.prototype.count = function () {
        return this.resultsdirty && (this.resultdata = this.resultset.data()), this.resultset.count();
      }, x.prototype.data = function (t) {
        return (this.sortDirty || this.resultsdirty) && this.performSortPhase({
          suppressRebuildEvent: !0
        }), this.options.persistent ? this.resultdata : this.resultset.data(t);
      }, x.prototype.queueRebuildEvent = function () {
        if (!this.rebuildPending) {
          this.rebuildPending = !0;
          var t = this;
          setTimeout(function () {
            t.rebuildPending && (t.rebuildPending = !1, t.emit("rebuild", t));
          }, this.options.minRebuildInterval);
        }
      }, x.prototype.queueSortPhase = function () {
        if (!this.sortDirty) {
          this.sortDirty = !0;
          var t = this;
          "active" === this.options.sortPriority ? setTimeout(function () {
            t.performSortPhase();
          }, this.options.minRebuildInterval) : this.queueRebuildEvent();
        }
      }, x.prototype.performSortPhase = function (t) {
        (this.sortDirty || this.resultsdirty) && (t = t || {}, this.sortDirty && (this.sortFunction ? this.resultset.sort(this.sortFunction) : this.sortCriteria ? this.resultset.compoundsort(this.sortCriteria) : this.sortCriteriaSimple && this.resultset.simplesort(this.sortCriteriaSimple.propname, this.sortCriteriaSimple.options), this.sortDirty = !1), this.options.persistent && (this.resultdata = this.resultset.data(), this.resultsdirty = !1), t.suppressRebuildEvent || this.emit("rebuild", this));
      }, x.prototype.evaluateDocument = function (t, e) {
        if (!this.resultset.filterInitialized) return this.options.persistent && (this.resultdata = this.resultset.data()), void (this.sortFunction || this.sortCriteria || this.sortCriteriaSimple ? this.queueSortPhase() : this.queueRebuildEvent());
        var i,
            r = this.resultset.filteredrows,
            n = e ? -1 : r.indexOf(+t),
            o = r.length,
            s = new O(this.collection);
        s.filteredrows = [t], s.filterInitialized = !0;

        for (var a = 0, l = this.filterPipeline.length; a < l; a++) i = this.filterPipeline[a], s[i.type](i.val);

        var h = 0 === s.filteredrows.length ? -1 : 0;
        return -1 !== n || -1 !== h ? -1 === n && -1 !== h ? (r.push(t), this.options.persistent && this.resultdata.push(this.collection.data[t]), void (this.sortFunction || this.sortCriteria || this.sortCriteriaSimple ? this.queueSortPhase() : this.queueRebuildEvent())) : -1 !== n && -1 === h ? (n < o - 1 ? (r.splice(n, 1), this.options.persistent && this.resultdata.splice(n, 1)) : (r.length = o - 1, this.options.persistent && (this.resultdata.length = o - 1)), void (this.sortFunction || this.sortCriteria || this.sortCriteriaSimple ? this.queueSortPhase() : this.queueRebuildEvent())) : -1 !== n && -1 !== h ? (this.options.persistent && (this.resultdata[n] = this.collection.data[t]), void (this.sortFunction || this.sortCriteria || this.sortCriteriaSimple ? this.queueSortPhase() : this.queueRebuildEvent())) : void 0 : void 0;
      }, x.prototype.removeDocument = function (t) {
        var e,
            i,
            r,
            n = {},
            o = {},
            s = [],
            a = this.resultset,
            l = this.resultset.filteredrows,
            h = l.length;
        if (!this.resultset.filterInitialized) return this.options.persistent && (this.resultdata = this.resultset.data()), void (this.sortFunction || this.sortCriteria || this.sortCriteriaSimple ? this.queueSortPhase() : this.queueRebuildEvent());

        for (Array.isArray(t) || (t = [t]), r = t.length, i = 0; i < r; i++) n[t[i]] = !0;

        for (e = 0; e < h; e++) n[l[e]] && (o[e] = !0);

        Object.keys(o).length > 0 && (this.resultset.filteredrows = this.resultset.filteredrows.filter(function (t, e) {
          return !o[e];
        }), this.options.persistent && (this.resultdata = this.resultdata.filter(function (t, e) {
          return !o[e];
        })), this.sortFunction || this.sortCriteria || this.sortCriteriaSimple ? this.queueSortPhase() : this.queueRebuildEvent());

        var c = function (t) {
          return function (e) {
            return e < a.filteredrows[t];
          };
        };

        for (h = a.filteredrows.length, e = 0; e < h; e++) s = t.filter(c(e)), a.filteredrows[e] -= s.length;
      }, x.prototype.mapReduce = function (t, e) {
        try {
          return e(this.data().map(t));
        } catch (t) {
          throw t;
        }
      }, k.prototype = new v(), k.prototype.createChange = function (t, e, i, r) {
        this.changes.push({
          name: t,
          operation: e,
          obj: "U" != e || this.disableDeltaChangesApi ? JSON.parse(JSON.stringify(i)) : this.getChangeDelta(i, r)
        });
      }, k.prototype.insertMeta = function (t) {
        var e, i;
        if (!this.disableMeta && t) if (Array.isArray(t)) for (e = t.length, i = 0; i < e; i++) t[i].hasOwnProperty("meta") || (t[i].meta = {}), t[i].meta.created = new Date().getTime(), t[i].meta.revision = 0;else t.meta || (t.meta = {}), t.meta.created = new Date().getTime(), t.meta.revision = 0;
      }, k.prototype.updateMeta = function (t) {
        !this.disableMeta && t && (t.meta.updated = new Date().getTime(), t.meta.revision += 1);
      }, k.prototype.createInsertChange = function (t) {
        this.createChange(this.name, "I", t);
      }, k.prototype.createUpdateChange = function (t, e) {
        this.createChange(this.name, "U", t, e);
      }, k.prototype.insertMetaWithChange = function (t) {
        this.insertMeta(t), this.createInsertChange(t);
      }, k.prototype.updateMetaWithChange = function (t, e) {
        this.updateMeta(t), this.createUpdateChange(t, e);
      }, k.prototype.console = {
        log: function () {},
        warn: function () {},
        error: function () {}
      }, k.prototype.addAutoUpdateObserver = function (t) {
        this.autoupdate && "function" == typeof Object.observe && Object.observe(t, this.observerCallback, ["add", "update", "delete", "reconfigure", "setPrototype"]);
      }, k.prototype.removeAutoUpdateObserver = function (t) {
        this.autoupdate && "function" == typeof Object.observe && Object.unobserve(t, this.observerCallback);
      }, k.prototype.addTransform = function (t, e) {
        if (this.transforms.hasOwnProperty(t)) throw new Error("a transform by that name already exists");
        this.transforms[t] = e;
      }, k.prototype.getTransform = function (t) {
        return this.transforms[t];
      }, k.prototype.setTransform = function (t, e) {
        this.transforms[t] = e;
      }, k.prototype.removeTransform = function (t) {
        delete this.transforms[t];
      }, k.prototype.byExample = function (t) {
        var e, i, r;

        for (e in r = [], t) t.hasOwnProperty(e) && r.push(((i = {})[e] = t[e], i));

        return {
          $and: r
        };
      }, k.prototype.findObject = function (t) {
        return this.findOne(this.byExample(t));
      }, k.prototype.findObjects = function (t) {
        return this.find(this.byExample(t));
      }, k.prototype.ttlDaemonFuncGen = function () {
        var t = this,
            e = this.ttl.age;
        return function () {
          var i = Date.now(),
              r = t.chain().where(function (t) {
            var r = t.meta.updated || t.meta.created,
                n = i - r;
            return e < n;
          });
          r.remove();
        };
      }, k.prototype.setTTL = function (t, e) {
        t < 0 ? clearInterval(this.ttl.daemon) : (this.ttl.age = t, this.ttl.ttlInterval = e, this.ttl.daemon = setInterval(this.ttlDaemonFuncGen(), e));
      }, k.prototype.prepareFullDocIndex = function () {
        for (var t = this.data.length, e = new Array(t), i = 0; i < t; i += 1) e[i] = i;

        return e;
      }, k.prototype.configureOptions = function (t) {
        (t = t || {}).hasOwnProperty("adaptiveBinaryIndices") && (this.adaptiveBinaryIndices = t.adaptiveBinaryIndices, this.adaptiveBinaryIndices && this.ensureAllIndexes());
      }, k.prototype.ensureIndex = function (t, e) {
        if (void 0 === e && (e = !1), null == t) throw new Error("Attempting to set index without an associated property");

        if ((!this.binaryIndices[t] || e || this.binaryIndices[t].dirty) && (!0 !== this.adaptiveBinaryIndices || !this.binaryIndices.hasOwnProperty(t) || e)) {
          var r = {
            name: t,
            dirty: !0,
            values: this.prepareFullDocIndex()
          };
          this.binaryIndices[t] = r;
          var n,
              o,
              s,
              a,
              l,
              h = (n = t, o = this.data, function (t, e) {
            if (~n.indexOf(".") ? (l = n.split("."), s = l.reduce(function (t, e) {
              return t && t[e] || void 0;
            }, o[t]), a = l.reduce(function (t, e) {
              return t && t[e] || void 0;
            }, o[e])) : (s = o[t][n], a = o[e][n]), s !== a) {
              if (i.lt(s, a, !1)) return -1;
              if (i.gt(s, a, !1)) return 1;
            }

            return 0;
          });
          r.values.sort(h), r.dirty = !1, this.dirty = !0;
        }
      }, k.prototype.checkAllIndexes = function (e) {
        var i,
            r = this.binaryIndices,
            n = [];

        for (i in r) t.call(r, i) && (this.checkIndex(i, e) || n.push(i));

        return n;
      }, k.prototype.checkIndex = function (t, e) {
        (e = e || {}).randomSamplingFactor && !1 !== e.randomSampling && (e.randomSampling = !0), e.randomSamplingFactor = e.randomSamplingFactor || .1, (e.randomSamplingFactor < 0 || e.randomSamplingFactor > 1) && (e.randomSamplingFactor = .1);
        var i,
            r,
            n,
            o,
            s,
            a = !0;
        if (!this.binaryIndices.hasOwnProperty(t)) throw new Error("called checkIndex on property without an index: " + t);
        if (this.adaptiveBinaryIndices || this.ensureIndex(t), s = this.binaryIndices[t].values, (o = s.length) !== this.data.length) return e.repair && this.ensureIndex(t, !0), !1;
        if (0 === o) return !0;
        if (1 === o) a = 0 === s[0];else if (e.randomSampling) {
          if (d.$lte(this.data[s[0]][t], this.data[s[1]][t]) || (a = !1), d.$lte(this.data[s[o - 2]][t], this.data[s[o - 1]][t]) || (a = !1), a) for (r = Math.floor((o - 1) * e.randomSamplingFactor), i = 0; i < r - 1; i++) if (n = Math.floor(Math.random() * (o - 1)), !d.$lte(this.data[s[n]][t], this.data[s[n + 1]][t])) {
            a = !1;
            break;
          }
        } else for (i = 0; i < o - 1; i++) if (!d.$lte(this.data[s[i]][t], this.data[s[i + 1]][t])) {
          a = !1;
          break;
        }
        return !a && e.repair && this.ensureIndex(t, !0), a;
      }, k.prototype.getBinaryIndexValues = function (t) {
        var e,
            i = this.binaryIndices[t].values,
            r = [];

        for (e = 0; e < i.length; e++) r.push(this.data[i[e]][t]);

        return r;
      }, k.prototype.ensureUniqueIndex = function (t) {
        var e = this.constraints.unique[t];
        return e || -1 == this.uniqueNames.indexOf(t) && this.uniqueNames.push(t), this.constraints.unique[t] = e = new N(t), this.data.forEach(function (t) {
          e.set(t);
        }), e;
      }, k.prototype.ensureAllIndexes = function (e) {
        var i,
            r = this.binaryIndices;

        for (i in r) t.call(r, i) && this.ensureIndex(i, e);
      }, k.prototype.flagBinaryIndexesDirty = function () {
        var e,
            i = this.binaryIndices;

        for (e in i) t.call(i, e) && (i[e].dirty = !0);
      }, k.prototype.flagBinaryIndexDirty = function (t) {
        this.binaryIndices[t] && (this.binaryIndices[t].dirty = !0);
      }, k.prototype.count = function (t) {
        return t ? this.chain().find(t).filteredrows.length : this.data.length;
      }, k.prototype.ensureId = function () {
        var t = this.data.length,
            e = 0;

        for (this.idIndex = []; e < t; e += 1) this.idIndex.push(this.data[e].$loki);
      }, k.prototype.ensureIdAsync = function (t) {
        this.async(function () {
          this.ensureId();
        }, t);
      }, k.prototype.addDynamicView = function (t, e) {
        var i = new x(this, t, e);
        return this.DynamicViews.push(i), i;
      }, k.prototype.removeDynamicView = function (t) {
        for (var e = 0; e < this.DynamicViews.length; e++) this.DynamicViews[e].name === t && this.DynamicViews.splice(e, 1);
      }, k.prototype.getDynamicView = function (t) {
        for (var e = 0; e < this.DynamicViews.length; e++) if (this.DynamicViews[e].name === t) return this.DynamicViews[e];

        return null;
      }, k.prototype.findAndUpdate = function (t, e) {
        "function" == typeof t ? this.updateWhere(t, e) : this.chain().find(t).update(e);
      }, k.prototype.findAndRemove = function (t) {
        this.chain().find(t).remove();
      }, k.prototype.insert = function (t) {
        if (!Array.isArray(t)) return this.insertOne(t);
        var e,
            i = [];
        this.emit("pre-insert", t);

        for (var r = 0, n = t.length; r < n; r++) {
          if (!(e = this.insertOne(t[r], !0))) return;
          i.push(e);
        }

        return this.emit("insert", i), 1 === (i = this.cloneObjects ? f(i, this.cloneMethod) : i).length ? i[0] : i;
      }, k.prototype.insertOne = function (t, e) {
        var i,
            r = null;
        if ("object" != typeof t ? r = new TypeError("Document needs to be an object") : null === t && (r = new TypeError("Object cannot be null")), null !== r) throw this.emit("error", r), r;
        var n = this.cloneObjects ? f(t, this.cloneMethod) : t;
        if (this.disableMeta || void 0 !== n.meta || (n.meta = {
          revision: 0,
          created: 0
        }), e || this.emit("pre-insert", n), this.add(n)) return this.disableChangesApi ? this.insertMeta(n) : this.insertMetaWithChange(n), i = this.cloneObjects ? f(n, this.cloneMethod) : n, e || this.emit("insert", i), this.addAutoUpdateObserver(i), i;
      }, k.prototype.clear = function (t) {
        var e = this;
        if (t = t || {}, this.data = [], this.idIndex = [], this.cachedIndex = null, this.cachedBinaryIndex = null, this.cachedData = null, this.maxId = 0, this.DynamicViews = [], this.dirty = !0, !0 === t.removeIndices) this.binaryIndices = {}, this.constraints = {
          unique: {},
          exact: {}
        }, this.uniqueNames = [];else {
          var i = Object.keys(this.binaryIndices);
          i.forEach(function (t) {
            e.binaryIndices[t].dirty = !1, e.binaryIndices[t].values = [];
          }), this.constraints = {
            unique: {},
            exact: {}
          }, this.uniqueNames.forEach(function (t) {
            e.ensureUniqueIndex(t);
          });
        }
      }, k.prototype.update = function (e) {
        var i, r, n;

        if (Array.isArray(e)) {
          n = e.length, (i = !this.cloneObjects && this.adaptiveBinaryIndices && Object.keys(this.binaryIndices).length > 0) && (this.adaptiveBinaryIndices = !1);

          try {
            for (r = 0; r < n; r += 1) this.update(e[r]);
          } finally {
            i && (this.ensureAllIndexes(), this.adaptiveBinaryIndices = !0);
          }
        } else {
          if (!t.call(e, "$loki")) throw new Error("Trying to update unsynced document. Please save the document first by using insert() or addMany()");

          try {
            this.startTransaction();
            var o,
                s,
                a,
                l,
                h,
                c = this.get(e.$loki, !0),
                u = this;
            if (!c) throw new Error("Trying to update a document not in collection.");
            o = c[0], a = c[1], s = this.cloneObjects || !this.disableDeltaChangesApi ? f(e, this.cloneMethod) : e, this.emit("pre-update", e), Object.keys(this.constraints.unique).forEach(function (t) {
              u.constraints.unique[t].update(o, s);
            }), this.data[a] = s, s !== e && this.addAutoUpdateObserver(e);

            for (var d = 0; d < this.DynamicViews.length; d++) this.DynamicViews[d].evaluateDocument(a, !1);

            if (this.adaptiveBinaryIndices) {
              var p = this.binaryIndices;

              for (l in p) this.adaptiveBinaryIndexUpdate(a, l);
            } else this.flagBinaryIndexesDirty();

            return this.idIndex[a] = s.$loki, this.commit(), this.dirty = !0, this.disableChangesApi ? this.updateMeta(s, null) : this.updateMetaWithChange(s, o), h = this.cloneObjects ? f(s, this.cloneMethod) : s, this.emit("update", h, o), h;
          } catch (t) {
            throw this.rollback(), this.console.error(t.message), this.emit("error", t), t;
          }
        }
      }, k.prototype.add = function (e) {
        if ("object" != typeof e) throw new TypeError("Object being added needs to be an object");
        if (void 0 !== e.$loki) throw new Error("Document is already in collection, please use update()");

        try {
          this.startTransaction(), this.maxId++, isNaN(this.maxId) && (this.maxId = this.data[this.data.length - 1].$loki + 1), e.$loki = this.maxId, this.disableMeta || (e.meta.version = 0);
          var i,
              r = this.constraints.unique;

          for (i in r) t.call(r, i) && r[i].set(e);

          this.idIndex.push(e.$loki), this.data.push(e);

          for (var n = this.data.length - 1, o = this.DynamicViews.length, s = 0; s < o; s++) this.DynamicViews[s].evaluateDocument(n, !0);

          if (this.adaptiveBinaryIndices) {
            var a = this.binaryIndices;

            for (i in a) this.adaptiveBinaryIndexInsert(n, i);
          } else this.flagBinaryIndexesDirty();

          return this.commit(), this.dirty = !0, this.cloneObjects ? f(e, this.cloneMethod) : e;
        } catch (t) {
          throw this.rollback(), this.console.error(t.message), this.emit("error", t), t;
        }
      }, k.prototype.updateWhere = function (t, e) {
        var i,
            r = this.where(t),
            n = 0;

        try {
          for (; n < r.length; n++) i = e(r[n]), this.update(i);
        } catch (t) {
          this.rollback(), this.console.error(t.message);
        }
      }, k.prototype.removeWhere = function (t) {
        var e;
        "function" == typeof t ? (e = this.data.filter(t), this.remove(e)) : this.chain().find(t).remove();
      }, k.prototype.removeDataOnly = function () {
        this.remove(this.data.slice());
      }, k.prototype.removeBatchByPositions = function (t) {
        var e,
            i,
            r,
            n,
            o = t.length,
            s = {},
            a = Object.keys(this.binaryIndices).length,
            l = Object.keys(this.constraints.unique).length,
            h = this.adaptiveBinaryIndices && Object.keys(this.binaryIndices).length > 0,
            c = this;

        try {
          for (this.startTransaction(), r = 0; r < o; r++) s[this.idIndex[t[r]]] = !0;

          if ((e = this.DynamicViews.length) > 0 || a > 0 || l > 0) {
            if (e > 0) for (i = 0; i < e; i++) this.DynamicViews[i].removeDocument(t);

            if (this.adaptiveBinaryIndices && !h) {
              var u,
                  d = this.binaryIndices;

              for (u in d) this.adaptiveBinaryIndexRemove(t, u);
            } else this.flagBinaryIndexesDirty();

            l && Object.keys(this.constraints.unique).forEach(function (e) {
              for (r = 0; r < o; r++) null !== (n = c.data[t[r]])[e] && void 0 !== n[e] && c.constraints.unique[e].remove(n[e]);
            });
          }

          if (!this.disableChangesApi || this.events.delete.length > 1) for (r = 0; r < o; r++) this.emit("delete", this.data[t[r]]);
          this.data = this.data.filter(function (t) {
            return !s[t.$loki];
          }), this.idIndex = this.idIndex.filter(function (t) {
            return !s[t];
          }), this.adaptiveBinaryIndices && h && (this.adaptiveBinaryIndices = !1, this.ensureAllIndexes(!0), this.adaptiveBinaryIndices = !0), this.commit(), this.dirty = !0;
        } catch (t) {
          return this.rollback(), h && (this.adaptiveBinaryIndices = !0), this.console.error(t.message), this.emit("error", t), null;
        }
      }, k.prototype.removeBatch = function (t) {
        var e,
            i = t.length,
            r = this.data.length,
            n = {},
            o = [];

        for (e = 0; e < r; e++) n[this.data[e].$loki] = e;

        for (e = 0; e < i; e++) "object" == typeof t[e] ? o.push(n[t[e].$loki]) : o.push(n[t[e]]);

        this.removeBatchByPositions(o);
      }, k.prototype.remove = function (e) {
        if ("number" == typeof e && (e = this.get(e)), "object" != typeof e) throw new Error("Parameter is not an object");
        if (Array.isArray(e)) this.removeBatch(e);else {
          if (!t.call(e, "$loki")) throw new Error("Object is not a document stored in the collection");

          try {
            this.startTransaction();
            var i = this.get(e.$loki, !0),
                r = i[1],
                n = this;
            Object.keys(this.constraints.unique).forEach(function (t) {
              null !== e[t] && void 0 !== e[t] && n.constraints.unique[t].remove(e[t]);
            });

            for (var o = 0; o < this.DynamicViews.length; o++) this.DynamicViews[o].removeDocument(r);

            if (this.adaptiveBinaryIndices) {
              var s,
                  a = this.binaryIndices;

              for (s in a) this.adaptiveBinaryIndexRemove(r, s);
            } else this.flagBinaryIndexesDirty();

            return this.data.splice(r, 1), this.removeAutoUpdateObserver(e), this.idIndex.splice(r, 1), this.commit(), this.dirty = !0, this.emit("delete", i[0]), delete e.$loki, delete e.meta, e;
          } catch (t) {
            return this.rollback(), this.console.error(t.message), this.emit("error", t), null;
          }
        }
      }, k.prototype.get = function (t, e) {
        var i = e || !1,
            r = this.idIndex,
            n = r.length - 1,
            o = 0,
            s = o + n >> 1;
        if (t = "number" == typeof t ? t : parseInt(t, 10), isNaN(t)) throw new TypeError("Passed id is not an integer");

        for (; r[o] < r[n];) r[s = o + n >> 1] < t ? o = s + 1 : n = s;

        return n === o && r[o] === t ? i ? [this.data[o], o] : this.data[o] : null;
      }, k.prototype.getBinaryIndexPosition = function (t, e) {
        var i = this.data[t][e],
            r = this.binaryIndices[e].values,
            n = this.calculateRange("$eq", e, i);
        if (0 === n[0] && -1 === n[1]) return null;

        for (var o = n[0], s = n[1], a = o; a <= s; a++) if (r[a] === t) return a;

        return null;
      }, k.prototype.adaptiveBinaryIndexInsert = function (t, e) {
        var i = this.binaryIndices[e].values,
            r = this.data[t][e];
        !0 === this.serializableIndices && r instanceof Date && (this.data[t][e] = r.getTime(), r = this.data[t][e]);
        var n = 0 === i.length ? 0 : this.calculateRangeStart(e, r, !0);
        this.binaryIndices[e].values.splice(n, 0, t);
      }, k.prototype.adaptiveBinaryIndexUpdate = function (t, e) {
        var i,
            r = this.binaryIndices[e].values,
            n = r.length;

        for (i = 0; i < n && r[i] !== t; i++);

        this.binaryIndices[e].values.splice(i, 1), this.adaptiveBinaryIndexInsert(t, e);
      }, k.prototype.adaptiveBinaryIndexRemove = function (t, e, i) {
        var r,
            n,
            o,
            s,
            a,
            l,
            h,
            c = this.binaryIndices[e],
            u = {};

        if (Array.isArray(t)) {
          if (1 !== (s = t.length)) {
            for (o = 0; o < s; o++) u[t[o]] = !0;

            if (c.values = c.values.filter(function (t) {
              return !u[t];
            }), !0 === i) return;
            var d = t.slice();

            for (d.sort(function (t, e) {
              return t - e;
            }), r = c.values.length, n = 0; n < r; n++) {
              for (a = c.values[n], l = 0, o = 0; o < s && a > d[o]; o++) l++;

              c.values[n] -= l;
            }

            return;
          }

          t = t[0];
        }

        if (null === (h = this.getBinaryIndexPosition(t, e))) return null;
        if (c.values.splice(h, 1), !0 !== i) for (r = c.values.length, n = 0; n < r; n++) c.values[n] > t && c.values[n]--;
      }, k.prototype.calculateRangeStart = function (t, e, r) {
        var n = this.data,
            o = this.binaryIndices[t].values,
            s = 0,
            a = o.length - 1,
            l = 0;
        if (0 === o.length) return -1;

        for (n[o[s]][t], n[o[a]][t]; s < a;) l = s + a >> 1, i.lt(n[o[l]][t], e, !1) ? s = l + 1 : a = l;

        var h = s;
        return i.aeq(e, n[o[h]][t]) ? h : i.lt(e, n[o[h]][t], !1) ? r ? h : h - 1 : r ? h + 1 : h;
      }, k.prototype.calculateRangeEnd = function (t, e) {
        var r = this.data,
            n = this.binaryIndices[t].values,
            o = 0,
            s = n.length - 1,
            a = 0;
        if (0 === n.length) return -1;

        for (r[n[o]][t], r[n[s]][t]; o < s;) a = o + s >> 1, i.lt(e, r[n[a]][t], !1) ? s = a : o = a + 1;

        var l = s;
        return i.aeq(e, r[n[l]][t]) ? l : i.gt(e, r[n[l]][t], !1) ? l + 1 : i.aeq(e, r[n[l - 1]][t]) ? l - 1 : l;
      }, k.prototype.calculateRange = function (t, e, r) {
        var n,
            o,
            s,
            a = this.data,
            l = this.binaryIndices[e].values,
            h = l.length - 1;
        if (0 === a.length) return [0, -1];
        var c = a[l[0]][e],
            u = a[l[h]][e];

        switch (t) {
          case "$eq":
          case "$aeq":
          case "$dteq":
            if (i.lt(r, c, !1) || i.gt(r, u, !1)) return [0, -1];
            break;

          case "$gt":
            if (i.gt(r, u, !0)) return [0, -1];
            if (i.gt(c, r, !1)) return [0, h];
            break;

          case "$gte":
            if (i.gt(r, u, !1)) return [0, -1];
            if (i.gt(c, r, !0)) return [0, h];
            break;

          case "$lt":
            if (i.lt(r, c, !0)) return [0, -1];
            if (i.lt(u, r, !1)) return [0, h];
            break;

          case "$lte":
            if (i.lt(r, c, !1)) return [0, -1];
            if (i.lt(u, r, !0)) return [0, h];
            break;

          case "$between":
            return i.gt(r[0], u, !1) ? [0, -1] : i.lt(r[1], c, !1) ? [0, -1] : (n = this.calculateRangeStart(e, r[0]), s = this.calculateRangeEnd(e, r[1]), n < 0 && n++, s > h && s--, i.gt(a[l[n]][e], r[0], !0) || n++, i.lt(a[l[s]][e], r[1], !0) || s--, s < n ? [0, -1] : [n, s]);

          case "$in":
            for (var d = [], p = [], f = 0, y = r.length; f < y; f++) for (var v = this.calculateRange("$eq", e, r[f]), g = v[0]; g <= v[1]; g++) void 0 === d[g] && (d[g] = !0, p.push(g));

            return p;
        }

        switch (t) {
          case "$eq":
          case "$aeq":
          case "$dteq":
          case "$gte":
          case "$lt":
            n = this.calculateRangeStart(e, r), o = a[l[n]][e];
        }

        switch (t) {
          case "$eq":
          case "$aeq":
          case "$dteq":
          case "$lte":
          case "$gt":
            s = this.calculateRangeEnd(e, r), a[l[s]][e];
        }

        switch (t) {
          case "$eq":
          case "$aeq":
          case "$dteq":
            return i.aeq(o, r) ? [n, s] : [0, -1];

          case "$gt":
            return i.aeq(a[l[s]][e], r) ? [s + 1, h] : [s, h];

          case "$gte":
            return i.aeq(a[l[n]][e], r) ? [n, h] : [n + 1, h];

          case "$lt":
            return i.aeq(a[l[n]][e], r) ? [0, n - 1] : [0, n];

          case "$lte":
            return i.aeq(a[l[s]][e], r) ? [0, s] : [0, s - 1];

          default:
            return [0, a.length - 1];
        }
      }, k.prototype.by = function (t, e) {
        var i;
        if (void 0 === e) return i = this, function (e) {
          return i.by(t, e);
        };
        var r = this.constraints.unique[t].get(e);
        return this.cloneObjects ? f(r, this.cloneMethod) : r;
      }, k.prototype.findOne = function (t) {
        t = t || {};
        var e = this.chain().find(t, !0).data();
        return Array.isArray(e) && 0 === e.length ? null : this.cloneObjects ? f(e[0], this.cloneMethod) : e[0];
      }, k.prototype.chain = function (t, e) {
        var i = new O(this);
        return void 0 === t ? i : i.transform(t, e);
      }, k.prototype.find = function (t) {
        return this.chain().find(t).data();
      }, k.prototype.findOneUnindexed = function (t, e) {
        for (var i = this.data.length; i--;) if (this.data[i][t] === e) return this.data[i];

        return null;
      }, k.prototype.startTransaction = function () {
        if (this.transactional) {
          this.cachedData = f(this.data, this.cloneMethod), this.cachedIndex = this.idIndex, this.cachedBinaryIndex = this.binaryIndices;

          for (var t = 0; t < this.DynamicViews.length; t++) this.DynamicViews[t].startTransaction();
        }
      }, k.prototype.commit = function () {
        if (this.transactional) {
          this.cachedData = null, this.cachedIndex = null, this.cachedBinaryIndex = null;

          for (var t = 0; t < this.DynamicViews.length; t++) this.DynamicViews[t].commit();
        }
      }, k.prototype.rollback = function () {
        if (this.transactional) {
          null !== this.cachedData && null !== this.cachedIndex && (this.data = this.cachedData, this.idIndex = this.cachedIndex, this.binaryIndices = this.cachedBinaryIndex);

          for (var t = 0; t < this.DynamicViews.length; t++) this.DynamicViews[t].rollback();
        }
      }, k.prototype.async = function (t, e) {
        setTimeout(function () {
          if ("function" != typeof t) throw new TypeError("Argument passed for async execution is not a function");
          t(), e();
        }, 0);
      }, k.prototype.where = function (t) {
        return this.chain().where(t).data();
      }, k.prototype.mapReduce = function (t, e) {
        try {
          return e(this.data.map(t));
        } catch (t) {
          throw t;
        }
      }, k.prototype.eqJoin = function (t, e, i, r, n) {
        return new O(this).eqJoin(t, e, i, r, n);
      }, k.prototype.stages = {}, k.prototype.getStage = function (t) {
        return this.stages[t] || (this.stages[t] = {}), this.stages[t];
      }, k.prototype.commitLog = [], k.prototype.stage = function (t, e) {
        var i = JSON.parse(JSON.stringify(e));
        return this.getStage(t)[e.$loki] = i, i;
      }, k.prototype.commitStage = function (t, e) {
        var i,
            r = this.getStage(t),
            n = new Date().getTime();

        for (i in r) this.update(r[i]), this.commitLog.push({
          timestamp: n,
          message: e,
          data: JSON.parse(JSON.stringify(r[i]))
        });

        this.stages[t] = {};
      }, k.prototype.no_op = function () {}, k.prototype.extract = function (t) {
        for (var e = 0, i = this.data.length, r = D(t), n = []; e < i; e += 1) n.push(q(this.data[e], t, r));

        return n;
      }, k.prototype.max = function (t) {
        return Math.max.apply(null, this.extract(t));
      }, k.prototype.min = function (t) {
        return Math.min.apply(null, this.extract(t));
      }, k.prototype.maxRecord = function (t) {
        for (var e, i = 0, r = this.data.length, n = D(t), o = {
          index: 0,
          value: void 0
        }; i < r; i += 1) void 0 !== e ? e < q(this.data[i], t, n) && (e = q(this.data[i], t, n), o.index = this.data[i].$loki) : (e = q(this.data[i], t, n), o.index = this.data[i].$loki);

        return o.value = e, o;
      }, k.prototype.minRecord = function (t) {
        for (var e, i = 0, r = this.data.length, n = D(t), o = {
          index: 0,
          value: void 0
        }; i < r; i += 1) void 0 !== e ? e > q(this.data[i], t, n) && (e = q(this.data[i], t, n), o.index = this.data[i].$loki) : (e = q(this.data[i], t, n), o.index = this.data[i].$loki);

        return o.value = e, o;
      }, k.prototype.extractNumerical = function (t) {
        return this.extract(t).map(S).filter(Number).filter(function (t) {
          return !isNaN(t);
        });
      }, k.prototype.avg = function (t) {
        return C(this.extractNumerical(t));
      }, k.prototype.stdDev = function (t) {
        return e = this.extractNumerical(t), i = C(e), r = C(e.map(function (t) {
          var e = t - i,
              r = e * e;
          return r;
        })), Math.sqrt(r);
        var e, i, r;
      }, k.prototype.mode = function (t) {
        var e,
            i,
            r,
            n = {},
            o = this.extract(t);

        for (i in o.forEach(function (t) {
          n[t] ? n[t] += 1 : n[t] = 1;
        }), n) e ? e < n[i] && (r = i) : (r = i, e = n[i]);

        return r;
      }, k.prototype.median = function (t) {
        var e = this.extractNumerical(t);
        e.sort(A);
        var i = Math.floor(e.length / 2);
        return e.length % 2 ? e[i] : (e[i - 1] + e[i]) / 2;
      }, j.prototype = {
        keys: [],
        values: [],
        sort: function (t, e) {
          return t < e ? -1 : t > e ? 1 : 0;
        },
        setSort: function (t) {
          this.bs = new E(t);
        },
        bs: function () {
          return new E(this.sort);
        },
        set: function (t, e) {
          var i = this.bs(this.keys, t);
          i.found ? this.values[i.index] = e : (this.keys.splice(i.index, 0, t), this.values.splice(i.index, 0, e));
        },
        get: function (t) {
          return this.values[$(this.keys, t, this.sort).index];
        }
      }, N.prototype.keyMap = {}, N.prototype.lokiMap = {}, N.prototype.set = function (t) {
        var e = t[this.field];

        if (null != e) {
          if (this.keyMap[e]) throw new Error("Duplicate key for property " + this.field + ": " + e);
          this.keyMap[e] = t, this.lokiMap[t.$loki] = e;
        }
      }, N.prototype.get = function (t) {
        return this.keyMap[t];
      }, N.prototype.byId = function (t) {
        return this.keyMap[this.lokiMap[t]];
      }, N.prototype.update = function (t, e) {
        if (this.lokiMap[t.$loki] !== e[this.field]) {
          var i = this.lokiMap[t.$loki];
          this.set(e), this.keyMap[i] = void 0;
        } else this.keyMap[t[this.field]] = e;
      }, N.prototype.remove = function (t) {
        var e = this.keyMap[t];
        if (null == e) throw new Error("Key is not in unique index: " + this.field);
        this.keyMap[t] = void 0, this.lokiMap[e.$loki] = void 0;
      }, N.prototype.clear = function () {
        this.keyMap = {}, this.lokiMap = {};
      }, M.prototype = {
        set: function (t, e) {
          this.index[t] ? this.index[t].push(e) : this.index[t] = [e];
        },
        remove: function (t, e) {
          var i = this.index[t];

          for (var r in i) i[r] == e && i.splice(r, 1);

          i.length < 1 && (this.index[t] = void 0);
        },
        get: function (t) {
          return this.index[t];
        },
        clear: function (t) {
          this.index = {};
        }
      }, g.LokiOps = d, g.Collection = k, g.KeyValueStore = j, g.LokiMemoryAdapter = m, g.LokiPartitioningAdapter = b, g.LokiLocalStorageAdapter = I, g.LokiFsAdapter = w, g.persistenceAdapters = {
        fs: w,
        localStorage: I
      }, g.aeq = n, g.lt = s, g.gt = a, g.Comparators = i, g;
    }();
  }),
      a = function () {
    function e() {
      t(this, e), this.queries = {}, this.db = void 0, this.load = this.load.bind(this);
    }

    return i(e, [{
      key: "load",
      value: function () {
        return new Promise(function (t, e) {
          var i = new s.persistenceAdapters.localStorage();
          void 0 !== window.indexedDB && (i = new o("client_analytics")), this.db = new s("games", {
            autosave: !0,
            autosaveInternal: 1e4,
            autoload: !0,
            autoloadCallback: function (i) {
              i ? e(i) : t();
            },
            adapter: i,
            env: "BROWSER"
          }), this.db.addCollection("events");
        }.bind(this));
      }
    }, {
      key: "registerQuery",
      value: function (t, e) {
        return this.queries[t] = e, Promise.resolve();
      }
    }, {
      key: "query",
      value: function (t, e, i) {
        return i = i || "events", Promise.resolve(this.queries[t](this.db.getCollection(i), e));
      }
    }, {
      key: "createCollection",
      value: function (t) {
        if ("events" === t) throw new Error("'events' is a reserved collection name");
        return this.db.addCollection(t), Promise.resolve();
      }
    }, {
      key: "insert",
      value: function (t, e) {
        var i = this.db.getCollection(t);
        if (void 0 === i) throw new Error(t + " is not a collection. Make sure to call createCollection first");
        return i.insert(e), Promise.resolve();
      }
    }, {
      key: "update",
      value: function (t, e) {
        var i = this.db.getCollection(t);
        if (void 0 === i) throw new Error(t + " is not a collection. Make sure to call createCollection first");
        return i.update(e), Promise.resolve();
      }
    }, {
      key: "remove",
      value: function (t, e) {
        var i = this.db.getCollection(t);
        if (void 0 === i) throw new Error(t + " is not a collection. Make sure to call createCollection first");
        return i.remove(e), Promise.resolve();
      }
    }]), e;
  }(),
      l = function () {
    function e() {
      t(this, e), this.queries = {}, this.bellhop = new window.Bellhop(), this.bellhop.connect();
    }

    return i(e, [{
      key: "load",
      value: function () {
        return Promise.resolve();
      }
    }, {
      key: "registerQuery",
      value: function (t, e) {
        return new Promise(function (i) {
          this.bellhop.fetch("client_analytics_registerQuery", i, {
            name: t,
            code: e.toString()
          });
        }.bind(this));
      }
    }, {
      key: "query",
      value: function (t, e, i) {
        return new Promise(function (r) {
          this.bellhop.fetch("client_analytics_query", function (t) {
            r(t.data);
          }, {
            name: t,
            params: e,
            collectionName: i
          });
        }.bind(this));
      }
    }, {
      key: "createCollection",
      value: function (t) {
        return new Promise(function (e) {
          this.bellhop.fetch("client_analytics_createCollection", e, {
            name: t
          });
        }.bind(this));
      }
    }]), e;
  }();

  l.prototype.insert = function (t, e) {
    return new Promise(function (i) {
      this.bellhop.fetch("client_analytics_insert", i, {
        collectionName: t,
        data: e
      });
    }.bind(this));
  }, l.prototype.update = function (t, e) {
    return new Promise(function (i) {
      this.bellhop.fetch("client_analytics_update", i, {
        collectionName: t,
        data: e
      });
    }.bind(this));
  }, l.prototype.remove = function (t, e) {
    return new Promise(function (i) {
      this.bellhop.fetch("client_analytics_remove", i, {
        collectionName: t,
        data: e
      });
    }.bind(this));
  }, "undefined" != typeof window && (window.PBS = window.PBS || {}, window.PBS.KIDS = window.PBS.KIDS || {}, window.PBS.KIDS.BellhopClientAnalytics = l);

  /**
   * @class BufferedQueue
   * @constructor
   * @param {string} localStorageKey The key for finding a queue in local storage
   * @param {string} eventServiceUrl The url of the LAP API receiving the events
   * @param {int} batchSize How many events get sent with each POST request
   * @param {int} flushRate How frequently (in ms) the auto-flush function should be called
   */

  var BufferedQueue =
  /*#__PURE__*/
  function () {
    function BufferedQueue(eventServiceUrl) {
      var _this = this;

      var batchSize = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 50;
      var flushRate = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : 2000;
      var collectionName = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : 'events';

      _classCallCheck(this, BufferedQueue);

      if (eventServiceUrl === undefined) {
        throw new Error('Missing Event Service URL in BufferedQueue constructor. Usage: new BufferedQueue(eventServiceUrl)');
      }

      this.eventServiceUrl = eventServiceUrl;
      this.batchSize = batchSize;
      this.flushRate = flushRate;
      this.collectionName = collectionName;
      /**
       * If we're online
       * @property {boolean} onlineStatus
       */

      this.onlineStatus = true;
      /**
       * Controls whether the buffer is calling its autoflush function
       */

      this.autoFlush = false;
      /**
       * Items waiting in queue
       * Checks if queue already exists in memory, and if not, creates empty array
       * @property {array} queuedItems
       */

      this.queuedItems = [];
      /**
       * Interval object that controls and calls the autoflush function
       */

      this.flushTimer = null;
      /**
       * If we're trying to get events from the database
       * @property {boolean} eventPolling
       */

      this.eventPolling = false;
      /**
       * Interval object that controls and calls the event polling function
       */

      this.pollingTimer = null; // before window closes, make sure to destroy (calls saveToStorage)

      window.addEventListener('beforeunload', function () {
        _this.destroyQueue();
      });
      /**
       * Client analytics database instance that holds event data when not in the queue
       */

      this.storageDb = new a(); // Initialize the database with our collection, two queries, and pull items that were queued but unsent

      this.ready = this.storageDb.load().then(function () {
        // check if collection specified exists, if not, create it
        var collection = _this.storageDb.db.getCollection(_this.collectionName);

        if (collection === undefined || collection === null) {
          _this.storageDb.createCollection(collectionName);
        } // pulls actual event data from db


        _this.storageDb.registerQuery('get_events', function (collection, params) {
          // split for whether or not you want to get a limited number of events
          if (params.limit) {
            return collection.chain().find({
              queueStatus: params.status
            }).limit(params.limit).simplesort('timestamp', false).data();
          } else {
            // getting all events of a particular type
            if (params.status) {
              return collection.chain().find({
                queueStatus: params.status
              }).simplesort('timestamp', false).data();
            } // getting all events in the database, period.
            else {
                return collection.chain().find(true).simplesort('timestamp', false).data();
              }
          }
        }); // gets just a count of the number of events with that status


        _this.storageDb.registerQuery('get_count', function (collection, params) {
          // count all events
          if (params.status === 'A') {
            return collection.find(true).length;
          } else {
            return collection.find({
              queueStatus: params.status
            }).length;
          }
        });
      }) // place 'in-queue' events back in the queue
      .then(function () {
        return _this.storageDb.query('get_events', {
          status: 'Q',
          limit: false
        }, _this.collectionName);
      }).then(function (oldQueue) {
        _this.queuedItems = oldQueue;
        return _this;
      });
    }
    /**
     * Wrapper for determining if queue has objects in it
     */


    _createClass(BufferedQueue, [{
      key: "isEmpty",
      value: function isEmpty() {
        if (this.queuedItems.length > 0) {
          return false;
        } else {
          return true;
        }
      }
      /**
       * Turn on the automatic getting of events from the database
       */

    }, {
      key: "enableEventPolling",
      value: function enableEventPolling() {
        this.eventPolling = true;
        this.pollingTimer = setInterval(this.populateQueue.bind(this), this.flushRate * 2);
      }
      /**
       * Turn off the automatic getting of events from the database
       */

    }, {
      key: "disableEventPolling",
      value: function disableEventPolling() {
        if (this.eventPolling) {
          this.eventPolling = false;
        }

        if (this.pollingTimer !== null) {
          clearInterval(this.pollingTimer);
          this.pollingTimer = null;
        }
      }
      /**
       * Stores event in backend database
       */

    }, {
      key: "storeEvent",
      value: function storeEvent(eventData) {
        var _this2 = this;

        this.runGarabageCollector().then(function () {
          eventData.queueStatus = 'U';
          return _this2.storageDb.insert(_this2.collectionName, eventData);
        });
      }
      /**
       * Removes event in backend database
       */

    }, {
      key: "deleteEvent",
      value: function deleteEvent(eventData) {
        return this.storageDb.remove(this.collectionName, eventData);
      }
      /**
       * In order to prevent the db from taking up too much memory,
       * on each insert, run a garbage collector to delete the oldest
       * sent events first, then old failed, then old unsent if no other choice.
       */

    }, {
      key: "runGarabageCollector",
      value: function runGarabageCollector() {
        var _this3 = this;

        return this.allEventsCount.then(function (count) {
          var toBeDeleted = false;

          var collection = _this3.storageDb.db.getCollection(_this3.collectionName);

          if (count > 200000) {
            // get oldest sent event
            toBeDeleted = _this3.getEventForDeletion(collection, 'S'); // if no sent events, get oldest failed event

            if (!toBeDeleted[0]) {
              toBeDeleted = _this3.getEventForDeletion(collection, 'F');
            } // if no sent or failed events, get oldest unsent event


            if (!toBeDeleted[0]) {
              toBeDeleted = _this3.getEventForDeletion(collection, 'U');
            }

            return _this3.deleteEvent(toBeDeleted[0]);
          } else {
            return Promise.resolve();
          }
        });
      }
      /**
       * Performs a search on the collection for the oldest item with the selected status
       * @param {Collection} collection The collection object that is being deleted from
       * @param {String} incomingStatus The status code being looked for
       */

    }, {
      key: "getEventForDeletion",
      value: function getEventForDeletion(collection, incomingStatus) {
        return collection.chain().find({
          queueStatus: incomingStatus
        }).simplesort('timestamp', false).limit(1).data();
      }
      /**
       * Generic queue status modifier function
       */

    }, {
      key: "markEventStatus",
      value: function markEventStatus(eventData, newStatus) {
        eventData.queueStatus = newStatus;
        return this.storageDb.update(this.collectionName, eventData);
      }
      /**
       * Put BATCH_SIZE * 2 events into the queue from the database, unsent > failed
       */

    }, {
      key: "populateQueue",
      value: function populateQueue() {
        var _this4 = this;

        return this.getEventsByStatus('U').then(function (events) {
          if (events.length === 0) {
            return _this4.getEventsByStatus('F');
          } else {
            return events;
          }
        }).then(function (events) {
          if (events.length !== 0) {
            events.forEach(function (event) {
              _this4.markEventStatus(event, 'Q');

              _this4.enqueueEvent(event);
            });
          }
        });
      }
      /**
       * Better reading wrapper for the db query
       */

    }, {
      key: "getEventsByStatus",
      value: function getEventsByStatus(incomingStatus) {
        return this.storageDb.query('get_events', {
          status: incomingStatus,
          limit: this.batchSize * 2
        }, this.collectionName);
      }
      /**
       * Better reading wrapper for the db query
       */

    }, {
      key: "getNumEventsByStatus",
      value: function getNumEventsByStatus(incomingStatus) {
        return this.storageDb.query('get_count', {
          status: incomingStatus,
          limit: false
        }, this.collectionName);
      }
      /**
       * Property wrappers for number of event getters
       */

    }, {
      key: "getSecondsToEmptyDb",
      value: function getSecondsToEmptyDb() {
        var _this5 = this;

        var unsentCount = this.unsentEventCount;
        var queuedCount = this.queuedEventCount;
        return Promise.all([unsentCount, queuedCount]).then(function (results) {
          var total = results[0] + results[1];
          return total / _this5.batchSize * (_this5.flushRate / 1000);
        });
      }
      /**
       * Add an event data object to queue
       * @param {object} eventData
       */

    }, {
      key: "enqueueEvent",
      value: function enqueueEvent(eventData) {
        this.queuedItems.push(eventData);
      }
      /**
       * Remove the next event data object from the queue
       */

    }, {
      key: "dequeueEvent",
      value: function dequeueEvent(newStatus) {
        if (!this.isEmpty()) {
          var poppedEvent = this.queuedItems.shift();
          return this.markEventStatus(poppedEvent, newStatus).then(function () {
            return poppedEvent;
          });
        } else {
          return false;
        }
      }
      /**
       * Sends event data to the API
       * @param {Array} batchOfEvents Event data to send
       */

    }, {
      key: "sendBatchToApi",
      value: function sendBatchToApi(batchOfEvents) {
        if (batchOfEvents.length > 0) {
          // need this b/c `this` will be different inside request object
          var queueLink = this;
          var request = new XMLHttpRequest();

          request.onload = function () {
            batchOfEvents.forEach(function (event) {
              queueLink.markEventStatus(event, 'S');
            });
          };

          request.onerror = function () {
            var _this6 = this;

            batchOfEvents.forEach(function (event) {
              if (_this6.status === 422) {
                // event was not validated, but was saved, so marked as sent
                queueLink.markEventStatus(event, 'S');
              } else {
                // event is marked as failed for later attempts at sending
                queueLink.markEventStatus(event, 'F');
              }
            });
          }; // mapping events to a copy so we can remove the loki metadata from the payload without screwing up the database


          var payload = batchOfEvents.map(function (event) {
            var copy = Object.assign({}, event);
            delete copy.meta;
            delete copy.$loki;
            return copy;
          });
          request.open('POST', this.eventServiceUrl);
          request.setRequestHeader('Content-Type', 'application/json');
          request.send(JSON.stringify(payload));
        }
      }
      /**
       * Creates an array of events from the queue, with the number controlled by batchSize
       */

    }, {
      key: "createBatch",
      value: function createBatch() {
        var batch = [];

        for (var i = 0; i < this.batchSize; i++) {
          if (!this.isEmpty()) {
            batch.push(this.dequeueEvent());
          } else {
            continue;
          }
        }

        return Promise.all(batch);
      }
      /**
       * Sets the onlineStatus variable and turns on polling
       * @param {boolean} onlineStatus
       */

    }, {
      key: "setOnline",
      value: function setOnline(onlineStatus) {
        this.onlineStatus = onlineStatus; // if turning from offline to online, begin polling for events from db and sending up to cloud

        if (onlineStatus) {
          this.enableEventPolling();
          this.enableAutoFlush();
        } // and the inverse
        else {
            this.disableEventPolling();
            this.disableAutoFlush();
          }
      }
      /**
       * Turns off autoflush and destroys the repeater
       */

    }, {
      key: "disableAutoFlush",
      value: function disableAutoFlush() {
        if (this.autoFlush) {
          this.autoFlush = false;
        }

        if (this.flushTimer !== null) {
          clearInterval(this.flushTimer);
          this.flushTimer = null;
        }
      }
      /**
       * Turns on autoflush and creates a repeater
       */

    }, {
      key: "enableAutoFlush",
      value: function enableAutoFlush() {
        this.autoFlush = true;
        this.flushTimer = setInterval(this.flushBatch.bind(this), this.flushRate);
      }
      /**
       * Sends one batch of data to the API, if the plugin is online
       */

    }, {
      key: "flushBatch",
      value: function flushBatch() {
        var _this7 = this;

        // dont want to try to flush if offline
        if (this.onlineStatus) {
          this.createBatch().then(function (batchOfEvents) {
            _this7.sendBatchToApi(batchOfEvents);
          });
        } else {
          return;
        }
      }
      /**
       * Calls the flushBatch function until the queue is empty
       */

    }, {
      key: "flushAll",
      value: function flushAll() {
        // dont want to try to flush if offline
        if (!this.onlineStatus) {
          return;
        }

        while (!this.isEmpty()) {
          this.flushBatch();
        }
      }
      /**
       * Just empties out the queue
       */

    }, {
      key: "clearQueue",
      value: function clearQueue() {
        while (!this.isEmpty()) {
          this.dequeueEvent('U');
        }
      }
      /**
       * Destroys the queue object so it is not active anymore
       */

    }, {
      key: "destroyQueue",
      value: function destroyQueue() {
        this.clearQueue();
        this.disableAutoFlush();
      }
    }, {
      key: "unsentEventCount",
      get: function get() {
        return this.getNumEventsByStatus('U');
      }
    }, {
      key: "failedEventCount",
      get: function get() {
        return this.getNumEventsByStatus('F');
      }
    }, {
      key: "queuedEventCount",
      get: function get() {
        return this.getNumEventsByStatus('Q');
      }
    }, {
      key: "sentEventCount",
      get: function get() {
        return this.getNumEventsByStatus('S');
      }
    }, {
      key: "allEventsCount",
      get: function get() {
        return this.getNumEventsByStatus('A');
      }
    }]);

    return BufferedQueue;
  }();

  /**
   * Utility class for detecting platform.
   * @class Platform
   */
  var Platform = function Platform() {
    _classCallCheck(this, Platform);

    /**
     * The name of the browser
     * @property {string} browser
     */
    this.browser = this.searchString(BROWSERS) || 'unknown browser';
    /**
     * The version of the user agent or app
     * @property {string} version
     */

    this.version = this.searchVersion(navigator.userAgent) || this.searchVersion(navigator.appVersion) || 'unknown version';
    /**
     * The version search string
     * @property {string} versionSearchString
     * @private
     */

    this.versionSearchString = null;
    /**
     * The current verison of the operating system
     * @property {string} OS
     */

    this.OS = this.searchString(OSES) || 'unknown OS';
  }; // reference to the prototype
  var p = Platform.prototype;
  /**
   * Search a string
   * @method searchString
   * @param {string|Array} data The collection of dat ato search
   * @return {null|String} The result or null
   */

  p.searchString = function (data) {
    for (var i = 0; i < data.length; i++) {
      var dataString = data[i].string;
      var dataProp = data[i].prop;
      this.versionSearchString = data[i].versionSearch || data[i].identity;

      if (dataString) {
        if (dataString.indexOf(data[i].subString) != -1) {
          return data[i].identity;
        }
      } else if (dataProp) {
        return data[i].identity;
      }
    }
  };
  /**
   *  Get the version for a string
   *  @method searchVersion
   *  @param {string} dataString The version to search for
   *  @return {null|number} The version number
   */


  p.searchVersion = function (dataString) {
    var index = dataString.indexOf(this.versionSearchString);
    if (index == -1) return;
    return parseFloat(dataString.substring(index + this.versionSearchString.length + 1));
  };
  /**
   *  The collection of BROWSERS types
   *  @property {array} BROWSERS
   *  @static
   *  @readOnly
   *  @private
   */


  var BROWSERS = [{
    string: navigator.userAgent,
    subString: 'Chrome',
    identity: 'Chrome'
  }, {
    string: navigator.userAgent,
    subString: 'OmniWeb',
    versionSearch: 'OmniWeb/',
    identity: 'OmniWeb'
  }, {
    string: navigator.vendor,
    subString: 'Apple',
    identity: 'Safari',
    versionSearch: 'Version'
  }, {
    prop: window.opera,
    identity: 'Opera',
    versionSearch: 'Version'
  }, {
    string: navigator.vendor,
    subString: 'iCab',
    identity: 'iCab'
  }, {
    string: navigator.vendor,
    subString: 'KDE',
    identity: 'Konqueror'
  }, {
    string: navigator.userAgent,
    subString: 'Firefox',
    identity: 'Firefox'
  }, {
    string: navigator.vendor,
    subString: 'Camino',
    identity: 'Camino'
  }, {
    // for newer Netscapes (6+)
    string: navigator.userAgent,
    subString: 'Netscape',
    identity: 'Netscape'
  }, {
    string: navigator.userAgent,
    subString: 'MSIE',
    identity: 'Explorer',
    versionSearch: 'MSIE'
  }, {
    string: navigator.userAgent,
    subString: 'Gecko',
    identity: 'Mozilla',
    versionSearch: 'rv'
  }, {
    // for older Netscapes (4-)
    string: navigator.userAgent,
    subString: 'Mozilla',
    identity: 'Netscape',
    versionSearch: 'Mozilla'
  }];
  /**
   *  The collection of OS types
   *  @property {array} OSES
   *  @static
   *  @readOnly
   *  @private
   */

  var OSES = [{
    string: navigator.platform,
    subString: 'Win',
    identity: 'WindowsPC'
  }, {
    string: navigator.platform,
    subString: 'Mac',
    identity: 'Mac'
  }, {
    string: navigator.userAgent,
    subString: 'iPhone',
    identity: 'iPhone/iPod'
  }, {
    string: navigator.userAgent,
    subString: 'iPad',
    identity: 'iPad'
  }, {
    string: navigator.userAgent,
    subString: 'Android',
    identity: 'Android'
  }, {
    string: navigator.platform,
    subString: 'Linux',
    identity: 'LinuxPC'
  }, {
    string: navigator.userAgent,
    subString: 'apple-phone-audio-bug',
    identity: 'iPhone/iPod'
  }, {
    string: navigator.userAgent,
    subString: 'apple-tablet-audio-bug',
    identity: 'iPad'
  }];

  var Identity;
  /**
   * A module that consumers can use to expose Super Vision kid labels to this plugin. Basically, the
   * "PBS.KIDS.CurrentKidLabel" (if it exists), should be an object with a get and set method to allow "setters" to
   * update the value on the fly, so that this plugin can always get the latest value when needed
   *
   * @var {Object} CurrentKidLabel
   */

  var CurrentKidLabel;
  /**
   * A module that consumers can use to expose the Super Vision channel id to this plugin. Basically, the
   * "PBS.KIDS.CurrentChannelId" (if it exists), should be an object witha get and set method to allow "setters" to
   * update the value on the fly, so that this plugin can always get the latest value when needed
   *
   * @var {Object} CurrentChannelId
   */

  var CurrentChannelId;
  /**
   * A module that consumers can use to expose the a content origin to this plugin. Basically, the
   * "PBS.KIDS.ContentOrigin" (if it exists), should be an object with a get and set method to allow "setters" to
   * update the value on the fly, so that this plugin can always get the latest value when needed. For instance, a value
   * might look like "org.pbskids.measureup", or "org.pbskids.gamesapp" to distinguish *where* the event came from. As
   * more Springroll games get embedded in multiple places, this will be valuable when wanting to filter those events
   * based on *where* the game was played (Measure Up vs. pbskids.org vs. Games vs. whatever)
   *
   * @var {Object} ContentOrigin
   */

  var ContentOrigin;
  /**
   * A module that consumers can use to expose a unique id for the installation of this app/game on a particular
   * device. Moving forward, Trampoline will provide this feature via Trampoline.util.getInstallationId(). However,
   * implementors can simply provide a global PBS.KIDS.InstallationId with a get method defined and this plugin should
   * pick it up and attach it to the event.
   */

  var InstallationId;
  /**
   * A module that consumers can use to expose the current study that the game is participating in. Implementors will
   * need this to be an object with a `get` method.
   */

  var StudyId;
  /**
   * A module that consumers can use to expose the current participant's id that's playing the game. Useful for studies.
   * Developers need to expose a global PBS.KIDS.ParticipantId.get() function that returns that value.
   */

  var ParticipantId;
  /**
   * Handle the Learning Analytics events
   * @class AnalyticsReporter
   * @constructor
   * @param {string} learningApiUrl The url for the LAP event collector end-point in the cloud
   */

  var AnalyticsReporter = function AnalyticsReporter(learningApiUrl) {
    _classCallCheck(this, AnalyticsReporter);

    var pbskids = {};

    if (window.PBS && window.PBS.KIDS) {
      pbskids = window.PBS.KIDS;
    }

    if (!learningApiUrl || learningApiUrl === '') {
      throw Error('learningApiUrl not set in AnalyticsReporter constructor. Check the container options, \
      and the constructor in the AnalyticsReporterPlugin setup() method.');
    }

    Identity = pbskids.identity;
    CurrentKidLabel = pbskids.CurrentKidLabel;
    CurrentChannelId = pbskids.CurrentChannelId;
    ContentOrigin = pbskids.ContentOrigin;
    InstallationId = pbskids.InstallationId;
    StudyId = pbskids.StudyId;
    ParticipantId = pbskids.ParticipantId;
    /**
     * The valid event keys for the resource type
     * @property {array} eventKeys
     * @private
     */

    this.eventKeys = EVENT_KEYS.v1;
    /**
     * Queuing object for the event buffer
     * @property {PBS.KIDS.BufferedQueue} queue
     */

    this.queue = new BufferedQueue(learningApiUrl, BATCH_SIZE, FLUSH_RATE, COLLECTION_NAME);
    /**
     * For getting platform details
     * @property {PBS.KIDS.Platform} platform
     */

    this.platform = new Platform();
    /**
     * The setInterval ping to check online status
     * @property {int} _timer
     * @private
     */

    this._timer = null;
    /**
     * Internal enabled boolean
     * @property {Boolean} _enabled
     * @private
     * @default true
     */

    this._enabled = false; // Enable the tracker by default

    this.enabled = false; // binding the current instance's status check so it can be removed from listeners in the destroy method

    this.checkStatus = this.checkStatus.bind(this);
    window.addEventListener('offline', this.checkStatus);
  };
  var COLLECTION_NAME = 'events';
  /**
   *  The time (in ms) between queue flushes
   *  @property {string} FLUSH_RATE
   *  @private
   *  @default 2000
   *  @static
   *  @readOnly
   */

  var FLUSH_RATE = 2000;
  /**
   *  The batch size of the queue
   *  @property {int} BATCH_SIZE
   *  @private
   *  @default 50
   *  @static
   *  @readOnly
   */

  var BATCH_SIZE = 50;
  /**
   *  JSON object keys for validation
   *  @property {object} EVENT_KEYS
   *  @private
   *  @static
   *  @readOnly
   */

  var EVENT_KEYS = {
    v2: {
      game: ['timestamp', 'user_ids', 'game_id', 'device_id', 'platform_id', 'event_id', 'event_data', 'game_session'],
      video: ['timestamp', 'user_ids', 'video_id', 'device_id', 'platform_id', 'event_id', 'event_data', 'video_session']
    },
    v1: ['timestamp', 'user_ids', 'game_id', 'device_id', 'platform_id', 'event_id', 'event_data', 'game_session']
  }; // Reference to the prototype

  var p$1 = AnalyticsReporter.prototype;
  /**
   * If the queueing should autoflush
   * @property {boolean} autoFlush
   */

  p$1.setAutoFlush = function (autoFlush) {
    if (autoFlush) {
      this.queue.enableAutoFlush();
    } else {
      this.queue.disableAutoFlush();
    }
  };
  /**
   * Check the online status internally on a timer
   * Auto calling function to keep updating online staus
   * @method checkStatus
   * @private
   */


  p$1.checkStatus = function () {
    if (!this.enabled) {
      return;
    }

    var queue = this.queue;
    isOnline().then(function (onlineStatus) {
      queue.setOnline(onlineStatus);
    });
  };
  /**
   * If the tracking is enabled
   * @property {boolean} enabled
   * @default true
   */


  Object.defineProperty(p$1, 'enabled', {
    get: function get() {
      return this._enabled;
    },
    set: function set(enabled) {
      this._enabled = enabled;

      if (this._timer) {
        clearInterval(this._timer);
        this._timer = null;
      }

      if (enabled) {
        // disable network on the buffer queue until we know we're online
        this.queue.setOnline(false); // perform an initial connectivity check

        this.checkStatus(); // schedule periodic connectivity checks

        this._timer = setInterval(this.checkStatus.bind(this), 10000);
        this.setAutoFlush(true);
      } else {
        this.setAutoFlush(false);
      }
    }
  });
  /**
   * Flush the queue
   * @method flushAll
   */

  p$1.flushAll = function () {
    this.queue.flushAll();
  };
  /**
   * Insert platform in JSON Data
   * @method appendPlatformDetails
   * @private
   * @param {object} eventData The event object data
   * @return {AnalyticsReporter} Instance for chaining
   */


  p$1.appendPlatformDetails = function (eventData) {
    eventData.platform_id = this.platform.browser;
    eventData.device_id = this.platform.OS;
    return this;
  };
  /**
   * Insert user information in JSON Data
   * @method appendUserData
   * @private
   * @param {object} eventData The event object data
   * @return {AnalyticsReporter} Instance for chaining
   */


  p$1.appendUserData = function (eventData) {
    if (Identity) {
      eventData.user_ids = [];
      var users = Identity.getCurrentUsers();

      for (var i = 0; i < users.length; i++) {
        if (users[i].userid) {
          eventData.user_ids.push(users[i].userid);
        }

        eventData.is_logged_in = users[i].isloggedin;
      }

      eventData.identities = Identity.getAllIdentities();
    } // If clients have defined a "PBS.KIDS.CurrentKidLabel" module, we'll also include the kid label there as well


    if (CurrentKidLabel instanceof Object) {
      eventData.kid_label_guid = CurrentKidLabel.get();
    }

    return this;
  };
  /**
   * Adds the current super vision channel id to the event JSON data
   * @method appendChannelId
   * @private
   * @param {object} eventData The event object data
   * @return {AnalyticsReporter} Instance for chaining
   */


  p$1.appendChannelId = function (eventData) {
    // If clients have defined a "PBS.KIDS.CurrentChannelId" module, we'll also include the channel id as well
    if (CurrentChannelId instanceof Object) {
      eventData.channel_id = CurrentChannelId.get();
    }

    return this;
  };
  /**
   * Adds the installation id for this app and device onto the event's JSON data, if the module is defined
   * @method appendInstallationId
   * @private
   * @param {object} eventData The event object data
   * @return {AnalyticsReporter} Instance for chaining
   */


  p$1.appendInstallationId = function (eventData) {
    if (InstallationId instanceof Object) {
      eventData.installation_id = InstallationId.get();
    }

    return this;
  };
  /**
   * Adds the currently configured study for this app, if the module is exposed AND there is a configured study
   * @method appendStudyId
   * @private
   * @param {object} eventData The event object data
   * @return {AnalyticsReporter} Instance for chaining
   */


  p$1.appendStudyId = function (eventData) {
    if (StudyId instanceof Object && StudyId.get() !== null) {
      eventData.study_id = StudyId.get();
    }

    return this;
  };

  p$1.appendParticipantId = function (eventData) {
    if (ParticipantId instanceof Object && ParticipantId.get() !== null) {
      eventData.participant_id = ParticipantId.get();
    }

    return this;
  };
  /**
   * Adds the current Super Vision content origin to the event JSON data
   * @method appendContentOrigin
   * @private
   * @param {Object} eventData The event object data
   * @return {AnalyticsReporter} Instance for chaining
   */


  p$1.appendContentOrigin = function (eventData) {
    // If clients have defined a content origin, we'll attach it
    if (ContentOrigin instanceof Object) {
      eventData.content_origin = ContentOrigin.get();
    }

    return this;
  };
  /**
   * Add the timestamp to the event JSON data
   * @method appendTimeStamp
   * @private
   * @param {object} eventData The event object data
   * @return {AnalyticsReporter} Instance for chaining
   */


  p$1.appendTimeStamp = function (eventData) {
    if (eventData.timestamp === undefined) {
      eventData.timestamp = Date.now();
    }

    return this;
  };
  /**
   * Push a learning event
   * @method pushEvent
   * @param {object} eventData The event object data
   * @param {string} eventData.game_id The GUID for the game
   * @param {string} eventData.event_id The GUID for the event
   * @param {object} eventData.event_data The data for event
   * @return {object} The event data with any appended data
   */


  p$1.pushEvent = function (eventData) {
    // Ignore if we aren't enabled
    if (!this._enabled) {
      return this;
    }

    if (eventData.user_ids === undefined) {
      eventData.user_ids = [];
    }

    eventData.is_logged_in = false;
    this.appendTimeStamp(eventData).appendPlatformDetails(eventData).appendUserData(eventData).appendChannelId(eventData).appendInstallationId(eventData).appendStudyId(eventData).appendParticipantId(eventData).appendContentOrigin(eventData);

    if (this.validateEvent(eventData)) {
      this.queue.storeEvent(eventData);
    }

    return eventData;
  };
  /**
   * Validate Event JSON data
   * @method validateEvent
   * @param {object} eventData
   * @private
   * @return {Boolean} if the event is valid
   */


  p$1.validateEvent = function (eventData) {
    for (var i = 0; i < this.eventKeys.length; i++) {
      if (!validateKey(eventData[this.eventKeys[i]])) {
        return false;
      }
    }

    return true;
  };
  /**
   * Validate event data keys from EVENT_KEYS
   * @method validateKaye
   * @private
   * @param {object} key Validate
   * @return {boolean} If the key is valid
   */


  var validateKey = function validateKey(key) {
    return !(key === null || key == 'undefined' || key === '');
  };
  /**
   * Cleanup and don't use after this
   * @method destroy
   */


  p$1.destroy = function () {
    this.enabled = false;
    this.queue.destroyQueue();
    this.queue = null;
    window.removeEventListener('offline', this.checkStatus);
    this.platform = null;
  };

  /**
   * Track an event for springroll Learning
   * @method onLearningEvent
   * @param {event} event The bellhop learningEvent
   * @private
   */

  var onLearningEvent = function onLearningEvent(event) {
    var data = event.data; // Send to the logging service if it's installed

    if (this.loggingService) {
      this.loggingService.send('pt-event', data);
    } // Automatically track with analytics


    this.analyticsReporter.pushEvent(data);
    /**
     * Event when dispatching a Learning Dispatcher event
     * @event learningEvent
     * @param {object} data The event data
     */

    this.trigger('learningEvent', data);
  };

  var AnalyticsReporterPlugin =
  /*#__PURE__*/
  function () {
    function AnalyticsReporterPlugin() {
      _classCallCheck(this, AnalyticsReporterPlugin);

      springroll.ContainerPlugin.call(this); // The Container Plugin constructor assigns the setup methods directly to the instance, so we have to overwrite
      // that with our methods

      this.setup = AnalyticsReporterPlugin.prototype.setup;
      this.open = AnalyticsReporterPlugin.prototype.open;
      this.close = AnalyticsReporterPlugin.prototype.close;
      this.teardown = AnalyticsReporterPlugin.prototype.teardown;
    }

    _createClass(AnalyticsReporterPlugin, [{
      key: "setup",
      value: function setup() {
        /**
         * The Learning Analytics client for sending events. Reads the learningApiUrl from the Container options
         * If the learningApiUrl is not set in the container options, this constructor will throw an error
         * @property {PBS.KIDS.AnalyticsReporter} analyticsReporter
         */
        this.analyticsReporter = new AnalyticsReporter(this.options.learningApiUrl);
      }
    }, {
      key: "open",
      value: function open() {
        // Enable the analytics dispatcher if we're in production mode
        this.analyticsReporter.enabled = true;
        var learningHandler = onLearningEvent.bind(this); // Client is setup, listen for learning event

        this.client.on('learningEvent', learningHandler);
      }
    }, {
      key: "close",
      value: function close() {
        if (this.analyticsReporter) {
          this.analyticsReporter.enabled = false;
          this.client.off('learningEvent');
        }
      }
    }, {
      key: "teardown",
      value: function teardown() {
        if (this.analyticsReporter) {
          this.analyticsReporter.destroy();
          delete this.analyticsReporter;
        }
      }
    }]);

    return AnalyticsReporterPlugin;
  }();

  // eslint-disable

  new AnalyticsReporterPlugin();

}());
